#! /usr/bin/env python
"""Checker for no-text-in-pdf."""
from __future__ import annotations

from pathlib import Path

import colrev.env.utils
import colrev.record.qm.quality_model
from colrev.constants import Fields
from colrev.constants import PDFDefectCodes

# pylint: disable=too-few-public-methods


class TextInPDFChecker:
    """The TextInPDFChecker"""

    msg = PDFDefectCodes.NO_TEXT_IN_PDF

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the no-text-in-pdf checks"""

        if (
            Fields.FILE not in record.data
            or Path(record.data[Fields.FILE]).suffix != ".pdf"
            or record.ignored_defect(key=Fields.FILE, defect=self.msg)
        ):
            return

        if not self._text_in_pdf(record=record):
            record.add_field_provenance_note(key=Fields.FILE, note=self.msg)
        else:
            record.remove_field_provenance_note(key=Fields.FILE, note=self.msg)

    def _text_in_pdf(
        self,
        *,
        record: colrev.record.record.Record,
    ) -> bool:
        if record.data[Fields.TEXT_FROM_PDF] == "":
            return False

        return True


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(TextInPDFChecker(quality_model))



#! /usr/bin/env python
"""Checker for title-not-in-pdf."""
from __future__ import annotations

import re
from pathlib import Path

import colrev.env.utils
import colrev.record.qm.quality_model
from colrev.constants import Fields
from colrev.constants import PDFDefectCodes

# pylint: disable=too-few-public-methods

# Note: replaces title_not_in_first_pages


class TitleNotInPDFChecker:
    """The TitleNotInPDFChecker"""

    msg = PDFDefectCodes.TITLE_NOT_IN_PDF

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the title-not-in-pdf checks"""

        if (
            Fields.FILE not in record.data
            or Path(record.data[Fields.FILE]).suffix != ".pdf"
            or Fields.TITLE not in record.data
            or record.data[Fields.TEXT_FROM_PDF] == ""
            or record.ignored_defect(key=Fields.FILE, defect=self.msg)
        ):
            return

        if not self._title_in_pdf(record=record):
            record.add_field_provenance_note(key=Fields.FILE, note=self.msg)
        else:
            record.remove_field_provenance_note(key=Fields.FILE, note=self.msg)

    def _title_in_pdf(self, *, record: colrev.record.record.Record) -> bool:
        text = record.data[Fields.TEXT_FROM_PDF]
        text = text.replace(" ", "").replace("\n", "").lower()
        text = colrev.env.utils.remove_accents(text)
        text = re.sub("[^a-zA-Z ]+", "", text)

        title_words = (
            re.sub("[^a-zA-Z ]+", "", record.data[Fields.TITLE]).lower().split()
        )

        match_count = 0
        for title_word in title_words:
            if title_word in text:
                match_count += 1

        if match_count / len(title_words) > 0.9:
            return True

        return False


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(TitleNotInPDFChecker(quality_model))



#! /usr/bin/env python
"""SearchSource: Crossref"""
from __future__ import annotations

import datetime
import typing
from multiprocessing import Lock
from pathlib import Path

import inquirer
import requests
import zope.interface
from pydantic import Field

import colrev.env.language_service
import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_settings
import colrev.packages.doi_org.src.doi_org as doi_connector
import colrev.record.record
import colrev.record.record_prep
import colrev.record.record_similarity
from colrev.constants import Fields
from colrev.constants import FieldValues
from colrev.constants import RecordState
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType
from colrev.packages.crossref.src import crossref_api


if typing.TYPE_CHECKING:  # pragma: no cover
    import colrev.settings

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class CrossrefSearchSource:
    """Crossref API"""

    endpoint = "colrev.crossref"
    source_identifier = Fields.DOI

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings
    search_types = [
        SearchType.API,
        SearchType.MD,
        SearchType.TOC,
    ]

    ci_supported: bool = Field(default=True)
    heuristic_status = SearchSourceHeuristicStatus.oni

    _api_url = "https://api.crossref.org/"

    def __init__(
        self,
        *,
        source_operation: colrev.process.operation.Operation,
        settings: typing.Optional[dict] = None,
    ) -> None:

        self.review_manager = source_operation.review_manager
        self.search_source = self._get_search_source(settings)
        self.crossref_lock = Lock()
        self.language_service = colrev.env.language_service.LanguageService()

        self._update_params()

        self.api = crossref_api.CrossrefAPI(params=self.search_source.search_parameters)

    def _update_params(self) -> None:

        if not self.search_source.search_parameters:
            # md-source
            return

        if "url" in self.search_source.search_parameters:
            return

        # convert params to crossref url
        if (
            "scope" in self.search_source.search_parameters
            and "issn" in self.search_source.search_parameters["scope"]
        ):

            url = (
                self._api_url
                + "journals/"
                + self.search_source.search_parameters["scope"]["issn"][0]
                + "/works"
            )
            self.search_source.search_parameters.pop("scope")

        elif "query" in self.search_source.search_parameters:
            url = (
                self._api_url
                + "works?"
                + "query.bibliographic="
                + self.search_source.search_parameters["query"].replace(" ", "+")
            )
            self.search_source.search_parameters.pop("query")

        else:
            raise NotImplementedError

        # pylint: disable=colrev-missed-constant-usage
        self.search_source.search_parameters["url"] = url
        self.search_source.search_parameters["version"] = "1.0.0"

        for position, source in enumerate(self.review_manager.settings.sources):
            if str(self.search_source.filename).endswith(str(source.filename)):
                self.review_manager.settings.sources[position] = self.search_source
                break

        self.review_manager.save_settings()

    def _get_search_source(
        self, settings: typing.Optional[dict]
    ) -> colrev.settings.SearchSource:
        if settings:
            # Crossref as a search_source
            return self.settings_class(**settings)

        # Crossref as an md-prep source
        crossref_md_filename = Path("data/search/md_crossref.bib")
        crossref_md_source_l = [
            s
            for s in self.review_manager.settings.sources
            if s.filename == crossref_md_filename
        ]
        if crossref_md_source_l:
            return crossref_md_source_l[0]

        return colrev.settings.SearchSource(
            endpoint="colrev.crossref",
            filename=crossref_md_filename,
            search_type=SearchType.MD,
            search_parameters={},
            comment="",
        )

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for Crossref"""

        result = {"confidence": 0.0}
        return result

    @classmethod
    def _parse_params(cls, params: str) -> dict:
        params_dict = {}
        if params:
            if params.startswith("http"):
                params_dict = {Fields.URL: params}
            else:
                for item in params.split(";"):
                    key, value = item.split("=")
                    params_dict[key] = value

        if "issn" in params_dict:
            params_dict["scope"] = {Fields.ISSN: [params_dict["issn"]]}  # type: ignore
            del params_dict["issn"]
        return params_dict

    @classmethod
    def _select_search_type(
        cls, operation: colrev.ops.search.Search, params_dict: dict
    ) -> SearchType:
        if list(params_dict) == ["scope"]:
            search_type = SearchType.TOC
        elif "query" in params_dict:
            search_type = SearchType.API
        elif Fields.URL in params_dict:
            search_type = SearchType.API
        else:
            search_type = operation.select_search_type(
                search_types=cls.search_types, params=params_dict
            )

        return search_type

    @classmethod
    def _add_toc_interactively(
        cls, *, operation: colrev.ops.search.Search
    ) -> colrev.settings.SearchSource:

        j_name = input("Enter journal name to lookup the ISSN:")

        endpoint = crossref_api.Endpoint(
            "https://api.crossref.org/journals?query=" + j_name.replace(" ", "+")
        )

        questions = [
            inquirer.List(
                Fields.JOURNAL,
                message="Select journal:",
                choices=[{x[Fields.TITLE]: x["ISSN"]} for x in endpoint],
            ),
        ]
        answers = inquirer.prompt(questions)
        issn = list(answers[Fields.JOURNAL].values())[0][0]

        filename = operation.get_unique_filename(f"crossref_issn_{issn}")
        add_source = colrev.settings.SearchSource(
            endpoint="colrev.crossref",
            filename=filename,
            search_type=SearchType.TOC,
            search_parameters={
                "url": f"https://api.crossref.org/journals/{issn.replace('-', '')}/works",
                "version": "1.0.0",
            },
            comment="",
        )
        return add_source

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint"""

        params_dict = cls._parse_params(params)
        search_type = cls._select_search_type(operation, params_dict)

        if search_type == SearchType.API:
            if len(params_dict) == 0:
                search_source = operation.create_api_source(endpoint=cls.endpoint)
                # pylint: disable=colrev-missed-constant-usage
                search_source.search_parameters["url"] = (
                    cls._api_url
                    + "works?"
                    + "query.bibliographic="
                    + search_source.search_parameters.pop("query", "").replace(" ", "+")
                )
                search_source.search_parameters["version"] = "1.0.0"
            else:
                if Fields.URL in params_dict:
                    query = {"url": params_dict[Fields.URL]}
                else:
                    query = params_dict

                filename = operation.get_unique_filename(file_path_string="crossref")
                search_source = colrev.settings.SearchSource(
                    endpoint="colrev.crossref",
                    filename=filename,
                    search_type=SearchType.API,
                    search_parameters=query,
                    comment="",
                )

        elif search_type == SearchType.TOC:
            if len(params_dict) == 0:
                search_source = cls._add_toc_interactively(operation=operation)
            else:
                filename = operation.get_unique_filename(file_path_string="crossref")
                search_source = colrev.settings.SearchSource(
                    endpoint="colrev.crossref",
                    filename=filename,
                    search_type=SearchType.TOC,
                    search_parameters=params_dict,
                    comment="",
                )

        else:
            raise NotImplementedError

        operation.add_source_and_search(search_source)
        return search_source

    def check_availability(
        self, *, source_operation: colrev.process.operation.Operation
    ) -> None:
        """Check status (availability) of the Crossref API"""
        self.api.check_availability(
            raise_service_not_available=(not self.review_manager.force_mode)
        )

    def _prep_crossref_record(
        self,
        *,
        record: colrev.record.record.Record,
        prep_main_record: bool = True,
        crossref_source: str = "",
    ) -> None:
        if Fields.LANGUAGE in record.data:
            try:
                self.language_service.unify_to_iso_639_3_language_codes(record=record)
            except colrev_exceptions.InvalidLanguageCodeException:
                del record.data[Fields.LANGUAGE]

        doi_connector.DOIConnector.get_link_from_doi(
            review_manager=self.review_manager,
            record=record,
        )

        if (
            self.review_manager.settings.is_curated_masterdata_repo()
        ) and Fields.CITED_BY in record.data:
            del record.data[Fields.CITED_BY]

        if not prep_main_record:
            # Skip steps for feed records
            return

        if FieldValues.RETRACTED in record.data.get(
            "warning", ""
        ) or FieldValues.RETRACTED in record.data.get(Fields.PRESCREEN_EXCLUSION, ""):
            record.prescreen_exclude(reason=FieldValues.RETRACTED)
            record.remove_field(key="warning")
        else:
            assert "" != crossref_source
            record.set_masterdata_complete(
                source=crossref_source,
                masterdata_repository=self.review_manager.settings.is_curated_repo(),
            )
            record.set_status(RecordState.md_prepared)

    def _validate_api_params(self) -> None:
        source = self.search_source

        if not all(x in ["url", "version"] for x in source.search_parameters):
            raise colrev_exceptions.InvalidQueryException(
                "Crossref search_parameters supports query or scope/issn field"
            )
        # Validate Query here

    def _validate_source(self) -> None:
        """Validate the SearchSource (parameters etc.)"""
        source = self.search_source
        self.review_manager.logger.debug(f"Validate SearchSource {source.filename}")

        if source.search_type not in self.search_types:
            raise colrev_exceptions.InvalidQueryException(
                f"Crossref search_type should be in {self.search_types}"
            )

        if source.search_type == SearchType.API:
            self._validate_api_params()

        self.review_manager.logger.debug(f"SearchSource {source.filename} validated")

    def _restore_url(
        self,
        *,
        record: colrev.record.record.Record,
        feed: colrev.ops.search_api_feed.SearchAPIFeed,
    ) -> None:
        """Restore the url from the feed if it exists
        (url-resolution is not always available)"""
        prev_record = feed.get_prev_feed_record(record)
        prev_url = prev_record.data.get(Fields.URL, None)
        if prev_url is None:
            return
        record.data[Fields.URL] = prev_url

    def _run_md_search(
        self,
        crossref_feed: colrev.ops.search_api_feed.SearchAPIFeed,
    ) -> None:

        for feed_record_dict in crossref_feed.feed_records.values():
            try:
                retrieved_record = self.api.query_doi(doi=feed_record_dict[Fields.DOI])

                if retrieved_record.data[Fields.DOI] != feed_record_dict[Fields.DOI]:
                    continue

                self._prep_crossref_record(
                    record=retrieved_record, prep_main_record=False
                )

                self._restore_url(record=retrieved_record, feed=crossref_feed)
                crossref_feed.add_update_record(retrieved_record)

            except (
                colrev_exceptions.RecordNotFoundInPrepSourceException,
                colrev_exceptions.NotFeedIdentifiableException,
            ):
                continue

        crossref_feed.save()

    def _scope_excluded(self, retrieved_record_dict: dict) -> bool:
        if (
            "scope" not in self.search_source.search_parameters
            or "years" not in self.search_source.search_parameters["scope"]
        ):
            return False

        year_from, year_to = self.search_source.search_parameters["scope"][
            "years"
        ].split("-")
        if not retrieved_record_dict.get(Fields.YEAR, -1000).isdigit():
            return True

        if (
            int(year_from)
            < int(retrieved_record_dict.get(Fields.YEAR, -1000))
            < int(year_to)
        ):
            return False
        return True

    def _potentially_overlapping_issn_search(self) -> bool:
        params = self.search_source.search_parameters
        if "scope" not in params:
            return False
        if Fields.ISSN not in params["scope"]:
            return False
        return len(params["scope"][Fields.ISSN]) > 1

    def _run_api_search(
        self,
        *,
        crossref_feed: colrev.ops.search_api_feed.SearchAPIFeed,
        rerun: bool,
    ) -> None:

        self.api.rerun = rerun
        self.api.last_updated = crossref_feed.get_last_updated()

        nrecs = self.api.get_len_total()
        self.review_manager.logger.info(f"Total: {nrecs:,} records")
        if not rerun:
            self.review_manager.logger.info(
                f"Retrieve papers indexed since {self.api.last_updated.split('T', maxsplit=1)[0]}"
            )
            nrecs = self.api.get_len()

        self.review_manager.logger.info(f"Retrieve {nrecs:,} records")
        estimated_time = nrecs * 0.5
        estimated_time_formatted = str(datetime.timedelta(seconds=int(estimated_time)))
        self.review_manager.logger.info(f"Estimated time: {estimated_time_formatted}")

        try:
            for retrieved_record in self.api.get_records():
                try:
                    if self._scope_excluded(retrieved_record.data):
                        continue

                    self._prep_crossref_record(
                        record=retrieved_record, prep_main_record=False
                    )

                    self._restore_url(record=retrieved_record, feed=crossref_feed)

                    crossref_feed.add_update_record(retrieved_record=retrieved_record)

                except colrev_exceptions.NotFeedIdentifiableException:
                    pass
        except RuntimeError as exc:
            print(exc)

        crossref_feed.save()

    def search(self, rerun: bool) -> None:
        """Run a search of Crossref"""

        self._validate_source()

        crossref_feed = self.search_source.get_api_feed(
            review_manager=self.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )

        if self.search_source.search_type in [
            SearchType.API,
            SearchType.TOC,
        ]:
            self._run_api_search(
                crossref_feed=crossref_feed,
                rerun=rerun,
            )
        elif self.search_source.search_type == SearchType.MD:
            self._run_md_search(crossref_feed)
        else:
            raise NotImplementedError

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.loader.load_utils.load(
                filename=self.search_source.filename,
                logger=self.review_manager.logger,
                unique_id_field="ID",
            )
            return records

        raise NotImplementedError

    def prepare(
        self,
        record: colrev.record.record_prep.PrepRecord,
        source: colrev.settings.SearchSource,
    ) -> colrev.record.record_prep.PrepRecord:
        """Source-specific preparation for Crossref"""
        source_item = [
            x
            for x in record.data[Fields.ORIGIN]
            if str(source.filename).replace("data/search/", "") in x
        ]
        if source_item:
            record.set_masterdata_complete(
                source=source_item[0],
                masterdata_repository=self.review_manager.settings.is_curated_repo(),
            )
        return record

    def _get_masterdata_record(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool,
    ) -> colrev.record.record.Record:
        try:
            try:
                retrieved_record = self.api.query_doi(doi=record.data[Fields.DOI])
            except (colrev_exceptions.RecordNotFoundInPrepSourceException, KeyError):

                retrieved_records = self.api.crossref_query(
                    record_input=record,
                    jour_vol_iss_list=False,
                )
                retrieved_record = retrieved_records.pop()

                retries = 0
                while (
                    not retrieved_record
                    and retries < prep_operation.max_retries_on_error
                ):
                    retries += 1

                    retrieved_records = self.api.crossref_query(
                        record_input=record,
                        jour_vol_iss_list=False,
                    )
                    retrieved_record = retrieved_records.pop()

            if not colrev.record.record_similarity.matches(record, retrieved_record):
                return record

            try:
                self.crossref_lock.acquire(timeout=120)

                # Note : need to reload file because the object is not shared between processes
                crossref_feed = self.search_source.get_api_feed(
                    review_manager=self.review_manager,
                    source_identifier=self.source_identifier,
                    update_only=False,
                    prep_mode=True,
                )

                crossref_feed.add_update_record(retrieved_record)

                record.merge(
                    retrieved_record,
                    default_source=retrieved_record.data[Fields.ORIGIN][0],
                )

                self._prep_crossref_record(
                    record=record,
                    crossref_source=retrieved_record.data[Fields.ORIGIN][0],
                )

                if save_feed:
                    crossref_feed.save()

            except colrev_exceptions.NotFeedIdentifiableException:
                pass
            finally:
                try:
                    self.crossref_lock.release()
                except ValueError:
                    pass

            return record

        except (
            colrev_exceptions.ServiceNotAvailableException,
            OSError,
            IndexError,
            colrev_exceptions.RecordNotFoundInPrepSourceException,
            colrev_exceptions.RecordNotParsableException,
        ) as exc:
            if prep_operation.review_manager.verbose_mode:
                print(exc)

        return record

    def _check_doi_masterdata(
        self, record: colrev.record.record.Record
    ) -> colrev.record.record.Record:
        try:
            retrieved_record = self.api.query_doi(doi=record.data[Fields.DOI])
            if not colrev.record.record_similarity.matches(record, retrieved_record):
                record.remove_field(key=Fields.DOI)

        except (
            requests.exceptions.RequestException,
            OSError,
            IndexError,
            colrev_exceptions.RecordNotFoundInPrepSourceException,
            colrev_exceptions.RecordNotParsableException,
        ):
            pass

        return record

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 30,
    ) -> colrev.record.record.Record:
        """Retrieve masterdata from Crossref based on similarity with the record provided"""

        # To test the metadata provided for a particular DOI use:
        # https://api.crossref.org/works/DOI
        if len(record.data.get(Fields.TITLE, "")) < 5 and Fields.DOI not in record.data:
            return record

        if Fields.DOI in record.data:
            record = self._check_doi_masterdata(record=record)

        record = self._get_masterdata_record(
            prep_operation=prep_operation,
            record=record,
            save_feed=save_feed,
        )

        return record



#!/usr/bin/env python3
"""Linter for CoLRev - missed constant usage"""
from __future__ import annotations

import typing

from astroid import Const
from astroid import nodes
from pylint import checkers
from pylint.checkers.utils import only_required_for_messages

from colrev.constants import DefectCodes
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.constants import FieldValues
from colrev.constants import OperationsType

if typing.TYPE_CHECKING:  # pragma: no cover
    from pylint.lint import PyLinter

# Should ensure that constants are used (instead of strings)


class MissedConstantUsageChecker(checkers.BaseChecker):
    """MissedConstantUsageChecker"""

    name = "colrev-missed-constant-usage"

    msgs = {
        "W0932": (
            "Missed constant usage",
            "colrev-missed-constant-usage",
            "Emitted when a string is used instead of a constant "
            + "(in colrev/constants.py) ",
        ),
    }

    constant_keys = [getattr(Fields, v) for v in dir(Fields) if not v.startswith("__")]
    constant_values = (
        [getattr(ENTRYTYPES, v) for v in dir(ENTRYTYPES) if not v.startswith("__")]
        + [getattr(FieldValues, v) for v in dir(FieldValues) if not v.startswith("__")]
        + [getattr(DefectCodes, v) for v in dir(DefectCodes) if not v.startswith("__")]
        + [
            getattr(OperationsType, v)
            for v in dir(OperationsType)
            if not v.startswith("__")
        ]
    )

    @only_required_for_messages("colrev-missed-constant-usage")
    def visit_assign(self, node: nodes.Assign) -> None:
        """
        Detect missed constant usage.
        """

        if len(node.targets) != 1:
            return

        assigned = node.value
        if isinstance(assigned, Const):
            if assigned.value in self.constant_values:
                self.add_message(self.name, node=node)  # , confidence=HIGH)

        if not hasattr(node.targets[0], "slice"):
            return

        if not hasattr(node.targets[0].slice, "value"):
            return

        if node.targets[0].slice.value in self.constant_keys:
            self.add_message(self.name, node=node)  # , confidence=HIGH)


def register(linter: PyLinter) -> None:  # pragma: no cover
    """required method to auto register this checker"""

    linter.register_checker(MissedConstantUsageChecker(linter))



#! /usr/bin/env python
"""Checker for name-format-separators fields."""
from __future__ import annotations

import re

import colrev.env.utils
import colrev.record.qm.quality_model
from colrev.constants import DefectCodes
from colrev.constants import Fields
from colrev.constants import FieldValues

# pylint: disable=too-few-public-methods


class NameFormatSeparatorsChecker:
    """The NameFormatSeparatorsChecker"""

    msg = DefectCodes.NAME_FORMAT_SEPARTORS

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the name-format-separators checks"""
        for key in [Fields.AUTHOR, Fields.EDITOR]:
            if (
                key not in record.data
                or record.ignored_defect(key=key, defect=self.msg)
                or record.data[key] == FieldValues.UNKNOWN
            ):
                continue

            if record.masterdata_is_curated():
                continue

            if self._name_separator_error(record=record, key=key):
                record.add_field_provenance_note(key=key, note=self.msg)
            else:
                record.remove_field_provenance_note(key=key, note=self.msg)

    def _name_separator_error(
        self, *, record: colrev.record.record.Record, key: str
    ) -> bool:

        # this should also catch organizations
        if "," not in record.data[key]:
            return True

        santized_names = colrev.env.utils.remove_accents(record.data[key])
        sanitized_names = re.sub(r"[{}]|\(\w*\b\)|\"\w*\"", "", santized_names)
        sanitized_names = sanitized_names.replace(" and others", "").replace(
            " et al.", ""
        )
        sanitized_names_list = sanitized_names.split(" and ")
        if not all(
            re.findall(
                r"^[\w .‐'’-]*, [\w .‐'’-]*$",
                sanitized_name,
                re.UNICODE,
            )
            for sanitized_name in sanitized_names_list
        ):
            return True

        # At least one upper case letter per name part
        if not all(
            any(char.isupper() for char in name_part)
            for sanitized_name in sanitized_names_list
            for name_part in sanitized_name.split(",")
        ):
            return True

        return False


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(NameFormatSeparatorsChecker(quality_model))



#! /usr/bin/env python
"""SearchSource: arXiv"""
from __future__ import annotations

import typing
from multiprocessing import Lock
from pathlib import Path
from urllib.parse import urlparse

import feedparser
import requests
import zope.interface
from pydantic import Field

import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
import colrev.record.record_prep
from colrev.constants import Fields
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType
from colrev.packages.arxiv.src import record_transformer

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class ArXivSource:
    """arXiv"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings

    endpoint = "colrev.arxiv"
    source_identifier = "arxivid"
    search_types = [SearchType.API]
    api_search_supported = True
    ci_supported: bool = Field(default=True)
    heuristic_status = SearchSourceHeuristicStatus.supported
    db_url = "https://arxiv.org/"
    _arxiv_md_filename = Path("data/search/md_arxiv.bib")

    def __init__(
        self,
        *,
        source_operation: colrev.process.operation.Operation,
        settings: typing.Optional[dict] = None,
    ) -> None:
        self.review_manager = source_operation.review_manager
        if settings:
            # arXiv as a search_source
            self.search_source = self.settings_class(**settings)
        else:
            # arXiv as an md-prep source
            arxiv_md_source_l = [
                s
                for s in self.review_manager.settings.sources
                if s.filename == self._arxiv_md_filename
            ]
            if arxiv_md_source_l:
                self.search_source = arxiv_md_source_l[0]
            else:
                self.search_source = colrev.settings.SearchSource(
                    endpoint="colrev.arxiv",
                    filename=self._arxiv_md_filename,
                    search_type=SearchType.API,
                    search_parameters={},
                    comment="",
                )

            self.arxiv_lock = Lock()

        self.operation = source_operation
        self.quality_model = self.review_manager.get_qm()
        _, self.email = self.review_manager.get_committer()

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for ArXiv"""

        result = {"confidence": 0.0}

        return result

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""

        params_dict = {}
        if params:
            if params.startswith("http"):
                params_dict = {Fields.URL: params}
            else:
                for item in params.split(";"):
                    key, value = item.split("=")
                    params_dict[key] = value

        # Note : always API search
        if len(params_dict) == 0:
            search_source = operation.create_api_source(endpoint=cls.endpoint)

        # pylint: disable=colrev-missed-constant-usage
        else:
            host = urlparse(params_dict["url"]).hostname

            assert host and host.endswith("arxiv.org")

            query = params_dict["url"].replace("https://arxiv.org/search/?query=", "")
            query = query[: query.find("&searchtype")]

            filename = operation.get_unique_filename(file_path_string="arxiv")

            search_source = colrev.settings.SearchSource(
                endpoint="colrev.arxiv",
                filename=filename,
                search_type=SearchType.API,
                search_parameters={"query": query},
                comment="",
            )

        operation.add_source_and_search(search_source)
        return search_source

    def validate_source(
        self,
        search_operation: colrev.ops.search.Search,
        source: colrev.settings.SearchSource,
    ) -> None:
        """Validate the SearchSource (parameters etc.)"""

        search_operation.review_manager.logger.debug(
            f"Validate SearchSource {source.filename}"
        )

        if source.filename.name != self._arxiv_md_filename.name:
            if "query" not in source.search_parameters:
                raise colrev_exceptions.InvalidQueryException(
                    f"Source missing query search_parameter ({source.filename})"
                )

            # if "query_file" in source.search_parameters:
            # ...

        search_operation.review_manager.logger.debug(
            f"SearchSource {source.filename} validated"
        )

    def check_availability(
        self, *, source_operation: colrev.process.operation.Operation
    ) -> None:
        """Check status (availability) of the ArXiv API"""

        # try:
        #     # pylint: disable=duplicate-code
        #     test_rec = {
        #         "author": "Wang, R E and Demszky, D ",
        #         "title": "Is ChatGPT a Good Teacher Coach?"
        #         "Measuring Zero-Shot Performance For Scoring and Providing "
        #           + \ "Actionable Insights on Classroom Instruction ",
        #         "ENTRYTYPE": "article",  # might not be needed in ArXiv
        #         "arxivid": "arXiv:2306.03090",
        #     }
        #     returned_record_dict = self._arxiv_query_id(
        #         arxiv_id=test_rec["arxivid"],
        #         timeout=20,
        #     )

        #     if returned_record_dict:
        #         assert returned_record_dict["title"] == test_rec["title"]
        #         assert returned_record_dict["author"] == test_rec["author"]
        #     else:
        #         if not source_operation.force_mode:
        #             raise colrev_exceptions.ServiceNotAvailableException("ArXiv")
        # except (requests.exceptions.RequestException, IndexError) as exc:
        #     print(exc)
        #     if not source_operation.force_mode:
        #         raise colrev_exceptions.ServiceNotAvailableException("ArXiv") from exc

    # def _arxiv_query_id(
    #     self,
    #     *,
    #     arxiv_id: str,
    #     timeout: int = 60,
    # ) -> dict:
    #     """Retrieve records from ArXiv based on a query"""

    #     # Query using ID List prefix ?? - wo hast du das gefunden?
    #     try:
    #         prefix = "id_list"
    #         url = (
    #             "https://export.arxiv.org/api/query?search_query="
    #             + f"list={prefix}:&id={arxiv_id}"
    #         )

    #         headers = {"user-agent": f"{__name__} (mailto:{self.email})"}
    #         session = self.review_manager.get_cached_session()

    #         # review_manager.logger.debug(url)
    #         ret = session.request("GET", url, headers=headers, timeout=timeout)
    #         ret.raise_for_status()
    #         if ret.status_code != 200:
    #             # review_manager.logger.debug(
    #             #     f"crossref_query failed with status {ret.status_code}"
    #             # )
    #             return {"arxiv_id": arxiv_id}

    #         input(str.encode(ret.text))
    #         root = fromstring(str.encode(ret.text))
    #         retrieved_record = self._arxiv_xml_to_record(root=root)
    #         if not retrieved_record:
    #             return {"arxiv_id": arxiv_id}
    #     except requests.exceptions.RequestException:
    #         return {"arxiv_id": arxiv_id}
    #     # pylint: disable=duplicate-code
    #     except OperationalError as exc:
    #         raise colrev_exceptions.ServiceNotAvailableException(
    #             "sqlite, required for requests CachedSession "
    #             "(possibly caused by concurrent operations)"
    #         ) from exc

    #     return retrieved_record

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Retrieve masterdata fromArXiv based on similarity with the record provided"""
        # https://info.arxiv.org/help/api/user-manual.html#_query_interface
        # id_list
        return record

    def _get_arxiv_ids(self, query: str, retstart: int) -> typing.List[dict]:
        url = (
            "https://export.arxiv.org/api/query?search_query="
            + f"all:{query}&start={retstart}&max_results=20"
        )
        feed = feedparser.parse(url)
        return feed["entries"]

    def _get_arxiv_query_return(self) -> typing.Iterator[dict]:
        params = self.search_source.search_parameters
        retstart = 0
        while True:
            entries = self._get_arxiv_ids(query=params["query"], retstart=retstart)
            if not entries:
                break
            for entry in entries:
                yield record_transformer.parse_record(entry)

            retstart += 20

    def _run_parameter_search(
        self,
        *,
        arxiv_feed: colrev.ops.search_api_feed.SearchAPIFeed,
        rerun: bool,
    ) -> None:
        if rerun:
            self.review_manager.logger.info(
                "Performing a search of the full history (may take time)"
            )

        try:
            for record_dict in self._get_arxiv_query_return():
                try:
                    # Note : discard "empty" records
                    if "" == record_dict.get(
                        Fields.AUTHOR, ""
                    ) and "" == record_dict.get(Fields.TITLE, ""):
                        self.review_manager.logger.warning(
                            f"Skipped record: {record_dict}"
                        )
                        continue

                    prep_record = colrev.record.record_prep.PrepRecord(record_dict)

                    added = arxiv_feed.add_update_record(prep_record)

                    # Note : only retrieve/update the latest deposits (unless in rerun mode)
                    if not added and not rerun:
                        # problem: some publishers don't necessarily
                        # deposit papers chronologically
                        break
                except colrev_exceptions.NotFeedIdentifiableException:
                    continue

            arxiv_feed.save()

        except requests.exceptions.JSONDecodeError as exc:
            # watch github issue:
            # https://github.com/fabiobatalha/crossrefapi/issues/46
            if "504 Gateway Time-out" in str(exc):
                raise colrev_exceptions.ServiceNotAvailableException(
                    "Crossref (check https://status.crossref.org/)"
                )
            raise colrev_exceptions.ServiceNotAvailableException(
                f"Crossref (check https://status.crossref.org/) ({exc})"
            )

    # def _run_md_search_update(
    #     self,
    #     *,
    #     arxiv_feed: colrev.ops.search.SearchAPIFeed,
    # ) -> None:
    #     records = self.review_manager.dataset.load_records_dict()

    #     for feed_record_dict in arxiv_feed.feed_records.values():
    #         feed_record = colrev.record.record.Record(feed_record_dict)

    #         try:
    #             retrieved_record = self._arxiv_query_id(
    #                 arxiv_id=feed_record_dict["arxivid"]
    #             )

    #             if retrieved_record["arxivid"] != feed_record.data["arxivid"]:
    #                 continue

    # prev_record_dict_version = (
    #     dblp_feed.get_prev_feed_record(
    #         retrieved_record=feed_record
    #     )
    # )
    #         retrieved_record = colrev.record.record.Record(retrieved_record)
    #         arxiv_feed.add_update_record(retrieved_record)

    #         changed = self.operation.update_existing_record(
    #             records=records,
    #             record_dict=retrieved_record,
    #             prev_record_dict_version=prev_record_dict_version,
    #             source=self.search_source,
    #         )
    #         if changed:
    #             arxiv_feed.nr_changed += 1
    #         except (
    #             colrev_exceptions.RecordNotFoundInPrepSourceException,
    #             colrev_exceptions.NotFeedIdentifiableException,
    #         ):
    #             continue

    #     arxiv_feed.save()

    def search(self, rerun: bool) -> None:
        """Run a search of ArXiv"""

        arxiv_feed = self.search_source.get_api_feed(
            review_manager=self.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )

        # if self.search_source.search_type == SearchType.MD:
        #     self._run_md_search_update(
        #         arxiv_feed=arxiv_feed,
        #     )

        if self.search_source.search_type == SearchType.API:
            self._run_parameter_search(
                arxiv_feed=arxiv_feed,
                rerun=rerun,
            )

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        # for API-based searches
        if self.search_source.filename.suffix == ".bib":
            records = colrev.loader.load_utils.load(
                filename=self.search_source.filename,
                logger=self.review_manager.logger,
            )
            for record in records.values():
                record[Fields.INSTITUTION] = "ArXiv"
            return records

        raise NotImplementedError

    def load_fixes(
        self,
        load_operation: colrev.ops.load.Load,
        source: colrev.settings.SearchSource,
        records: typing.Dict,
    ) -> dict:
        """Load fixes for ArXiv"""
        return records

    def prepare(
        self, record: colrev.record.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.record.Record:
        """Source-specific preparation for ArXiv"""

        if Fields.AUTHOR in record.data:
            record.data[Fields.AUTHOR] = (
                colrev.record.record_prep.PrepRecord.format_author_field(
                    record.data[Fields.AUTHOR]
                )
            )

        return record



#! /usr/bin/env python
"""Structured data extraction as part of the data operations"""
from __future__ import annotations

import csv
import typing
from dataclasses import asdict
from pathlib import Path

import pandas as pd
import zope.interface
from git.exc import GitCommandError
from pydantic import BaseModel
from pydantic import Field

import colrev.env.utils
import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Colors
from colrev.constants import Fields
from colrev.constants import RecordState


# pylint: disable=too-few-public-methods


# an option: https://pypi.org/project/csv-schema/
class DataField(BaseModel):
    """Field definition"""

    name: str
    explanation: str
    data_type: str


class StructuredDataSettings(BaseModel):
    """Settings for StructuredData"""

    endpoint: str
    version: str
    fields: typing.List[DataField]
    data_path_relative: Path = Path("data.csv")

    _details = {
        "fields": {"tooltip": "Fields for the structured data extraction"},
    }


@zope.interface.implementer(colrev.package_manager.interfaces.DataInterface)
class StructuredData:
    """Summarize the literature in a structured data extraction (a table)"""

    settings: StructuredDataSettings
    settings_class = StructuredDataSettings

    _FULL_DATA_FIELD_EXPLANATION = """Explanation: Data fields are used in the coding tables.
    Example 1:
    - name           : summary
    - explanation    : Brief summary of the principal findings
    - data_type      : str

    Example 2:
    - name           : sample_size
    - explanation    : Sample size of the study
    - data_type      : int
    """
    ci_supported: bool = Field(default=False)

    def __init__(
        self,
        *,
        data_operation: colrev.ops.data.Data,
        settings: dict,
    ) -> None:
        self.review_manager = data_operation.review_manager

        if "version" not in settings:
            settings["version"] = "0.1"

        if "fields" not in settings:
            settings["fields"] = []
        if "data_path_relative" not in settings:
            settings["data_path_relative"] = Path("data/data/data.csv")

        self.settings = self.settings_class(**settings)
        self.data_path = self.review_manager.path / self.settings.data_path_relative
        self.review_manager = self.review_manager

    # pylint: disable=unused-argument
    @classmethod
    def add_endpoint(cls, operation: colrev.ops.data.Data, params: str) -> None:
        """Add as an endpoint"""

        add_source = {
            "endpoint": "colrev.structured",
            "version": "0.1",
            "fields": [],
            "data_path_relative": "data/data/data.csv",
        }
        operation.review_manager.settings.data.data_package_endpoints.append(add_source)
        operation.review_manager.save_settings()
        operation.review_manager.dataset.create_commit(
            msg=f"Add {operation.type} colrev.structured",
        )

    def validate_structured_data(self) -> None:
        """Validate the extracted data"""

        if not self.data_path.is_file():
            return

        data_df = pd.read_csv(self.data_path, dtype=str)

        # Check for duplicate IDs
        if not data_df[Fields.ID].is_unique:
            raise colrev_exceptions.DataException(
                msg=f"duplicates in {self.settings.data_path_relative}: "
                + ",".join(data_df[data_df.duplicated([Fields.ID])].ID.tolist())
            )

        # Check consistency: data -> inclusion_2
        data_ids = data_df[Fields.ID].tolist()
        records = self.review_manager.dataset.load_records_dict()
        for data_id in data_ids:
            if data_id not in records:
                raise colrev_exceptions.DataException(
                    msg=f"{data_id} in {self.settings.data_path_relative} not in records"
                )
            if records[data_id][Fields.STATUS] not in [
                RecordState.rev_included,
                RecordState.rev_synthesized,
            ]:
                raise colrev_exceptions.DataException(
                    msg=f"{data_id} in {self.settings.data_path_relative} "
                    + "not in [rev_included, rev_synthesized]"
                )

        # Note : missing IDs are added through update_data

    def _set_fields(self) -> None:
        self.review_manager.logger.info("Add fields for data extraction")
        try:
            _ = self.review_manager.dataset.get_repo()
        except GitCommandError:
            return

        print(self._FULL_DATA_FIELD_EXPLANATION)
        while "y" == input("Add a data field [y,n]?"):
            short_name = input("Provide a short name: ")
            explanation = input("Provide a short explanation: ")
            data_type = input("Provide the data type (str, int, double):")

            self.settings.fields.append(
                Field(
                    **{
                        "name": short_name,
                        "explanation": explanation,
                        "data_type": data_type,
                    }
                )
            )

            print()

        # Note : the following should be easier...
        for i, candidate in enumerate(
            self.review_manager.settings.data.data_package_endpoints
        ):
            if candidate.get("data_path_relative", "NA") == str(
                self.settings.data_path_relative
            ):
                self.review_manager.settings.data.data_package_endpoints[i] = asdict(
                    self.settings, dict_factory=colrev.env.utils.custom_asdict_factory
                )
        self.review_manager.save_settings()

    def update_data(
        self,
        records: dict,
        synthesized_record_status_matrix: dict,
        silent_mode: bool,
    ) -> None:
        """Update the data/structured data extraction"""

        def update_structured_data(
            *,
            review_manager: colrev.review_manager.ReviewManager,
            synthesized_record_status_matrix: dict,
        ) -> typing.Dict:
            if silent_mode:
                return synthesized_record_status_matrix
            if not self.data_path.is_file():
                self._set_fields()

                field_names = [f.name for f in self.settings.fields]
                data_df = pd.DataFrame([], columns=[Fields.ID] + field_names)
                data_df.sort_values(by=[Fields.ID], inplace=True)

                data_df.to_csv(self.data_path, index=False, quoting=csv.QUOTE_ALL)

            nr_records_added = 0

            if not silent_mode:
                self.review_manager.report_logger.info("Update structured data")
                self.review_manager.logger.info(
                    f"Update structured data ({self.settings.data_path_relative})"
                )

            data_df = pd.read_csv(self.data_path, dtype=str)

            for record_id in list(synthesized_record_status_matrix.keys()):
                # skip when already available
                if 0 < len(data_df[data_df[Fields.ID].str.startswith(record_id)]):
                    continue

                add_record = pd.DataFrame({Fields.ID: [record_id]})
                add_record = add_record.reindex(
                    columns=data_df.columns, fill_value="TODO"
                )
                data_df = pd.concat([data_df, add_record], axis=0, ignore_index=True)
                review_manager.logger.info(
                    f" {Colors.GREEN}{record_id}".ljust(45)
                    + f"add to structured_data{Colors.END}"
                )
                nr_records_added = nr_records_added + 1

            data_df.sort_values(by=[Fields.ID], inplace=True)

            data_df.to_csv(self.data_path, index=False, quoting=csv.QUOTE_ALL)

            if not (0 == nr_records_added and silent_mode):
                review_manager.logger.info(
                    f"Added to {self.settings.data_path_relative}".ljust(24)
                    + f"{nr_records_added}".rjust(15, " ")
                    + " records"
                )

                review_manager.logger.info(
                    f"Added to {self.settings.data_path_relative}".ljust(24)
                    + f"{nr_records_added}".rjust(15, " ")
                    + " records"
                )
            return records

        self.validate_structured_data()
        records = update_structured_data(
            review_manager=self.review_manager,
            synthesized_record_status_matrix=synthesized_record_status_matrix,
        )

        self.review_manager.dataset.add_changes(
            self.settings.data_path_relative, ignore_missing=True
        )

    def update_record_status_matrix(
        self,
        synthesized_record_status_matrix: dict,
        endpoint_identifier: str,
    ) -> None:
        """Update the record_status_matrix"""

        def get_data_extracted(
            data_path: Path, records_for_data_extraction: list
        ) -> list:
            data_extracted = []
            data_df = pd.read_csv(data_path)

            for record in records_for_data_extraction:
                drec = data_df.loc[data_df[Fields.ID] == record]
                if 1 == drec.shape[0]:
                    if "TODO" not in drec.iloc[0].tolist():
                        data_extracted.append(drec.loc[drec.index[0], Fields.ID])

            data_extracted = [
                x for x in data_extracted if x in records_for_data_extraction
            ]
            return data_extracted

        def get_structured_data_extracted(
            *, synthesized_record_status_matrix: typing.Dict, data_path: Path
        ) -> list:
            if not data_path.is_file():
                return []

            data_extracted = get_data_extracted(
                data_path, list(synthesized_record_status_matrix.keys())
            )
            data_extracted = [
                x
                for x in data_extracted
                if x in list(synthesized_record_status_matrix.keys())
            ]
            return data_extracted

        structured_data_extracted = get_structured_data_extracted(
            synthesized_record_status_matrix=synthesized_record_status_matrix,
            data_path=self.data_path,
        )

        for syn_id in structured_data_extracted:
            if syn_id in synthesized_record_status_matrix:
                synthesized_record_status_matrix[syn_id][endpoint_identifier] = True
            else:
                print(f"Error: {syn_id} not int " f"{synthesized_record_status_matrix}")

    def get_advice(
        self,
    ) -> dict:
        """Get advice on the next steps (for display in the colrev status)"""

        data_endpoint = "Data operation [structured data endpoint]: "

        if self.settings.data_path_relative.is_file():
            advice = {
                "msg": f"{data_endpoint}"
                + f"\n    - Complete the data extraction ({self.settings.data_path_relative})",
                "detailed_msg": "TODO",
            }
        else:
            advice = {}
        return advice



#! /usr/bin/env python
"""CoLRev search operation: Search for relevant records."""
from __future__ import annotations

import typing
from pathlib import Path

import inquirer

import colrev.exceptions as colrev_exceptions
import colrev.process.operation
import colrev.settings
from colrev.constants import Colors
from colrev.constants import EndpointType
from colrev.constants import Fields
from colrev.constants import OperationsType
from colrev.constants import SearchType
from colrev.writer.write_utils import write_file


class Search(colrev.process.operation.Operation):
    """Search for new records"""

    type = OperationsType.search

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        notify_state_transition_operation: bool = True,
    ) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=self.type,
            notify_state_transition_operation=notify_state_transition_operation,
        )
        self.review_manager = review_manager
        self.sources = review_manager.settings.sources
        self.package_manager = self.review_manager.get_package_manager()

    def get_unique_filename(self, file_path_string: str, suffix: str = ".bib") -> Path:
        """Get a unique filename for a (new) SearchSource"""

        self.review_manager.load_settings()
        self.sources = self.review_manager.settings.sources

        file_path_string = (
            file_path_string.replace("+", "_").replace(" ", "_").replace(";", "_")
        )

        if file_path_string.endswith(suffix):
            file_path_string = file_path_string.rstrip(suffix)
        filename = Path(f"data/search/{file_path_string}{suffix}")
        existing_filenames = [x.filename for x in self.sources]
        if all(x != filename for x in existing_filenames):
            return filename

        i = 1
        while not all(x != filename for x in existing_filenames):
            filename = Path(f"data/search/{file_path_string}_{i}{suffix}")
            i += 1

        return filename

    def get_query_filename(
        self, *, filename: Path, instantiate: bool = False, interactive: bool = False
    ) -> Path:
        """Get the corresponding filename for the search query"""
        query_filename = Path("data/search/") / Path(str(filename.stem) + "_query.txt")
        if instantiate:
            with open(query_filename, "w", encoding="utf-8") as file:
                file.write("")
            if interactive:
                input(
                    f"Created {Colors.ORANGE}{query_filename}{Colors.END}. "
                    "Please store your query in the file and press Enter to continue."
                )
            self.review_manager.dataset.add_changes(query_filename)
        return query_filename

    def create_db_source(  # type: ignore
        self, *, search_source_cls, params: dict
    ) -> colrev.settings.SearchSource:
        """Interactively add a DB SearchSource"""

        if not all(x in ["search_file"] for x in list(params)):
            raise NotImplementedError  # or parameter error

        if "search_file" in params:
            filename = Path(params["search_file"])
        else:
            filename = self.get_unique_filename(
                file_path_string=search_source_cls.endpoint.replace("colrev.", "")
            )
        self.review_manager.logger.debug(f"Add new DB source: {filename}")

        query_file = self.get_query_filename(filename=filename, instantiate=True)
        self.review_manager.logger.info(f"Created query-file: {query_file}")
        input(
            f"{Colors.ORANGE}Store query in query-file and press Enter to continue{Colors.END}"
        )

        if not filename.is_file():
            self.review_manager.logger.info(
                f"- Go to {Colors.ORANGE}{search_source_cls.db_url}{Colors.END}"
            )
            query_file = self.get_query_filename(
                filename=filename, instantiate=True, interactive=False
            )
            self.review_manager.logger.info(
                f"- Search for your query and store it in {Colors.ORANGE}{query_file}{Colors.END}"
            )
            self.review_manager.logger.info(
                f"- Save search results in {Colors.ORANGE}{filename}{Colors.END}"
            )
            input("Press Enter to complete")

        self.review_manager.dataset.add_changes(filename, ignore_missing=True)
        self.review_manager.dataset.add_changes(query_file)

        add_source = colrev.settings.SearchSource(
            endpoint=search_source_cls.endpoint,
            filename=filename,
            search_type=SearchType.DB,
            search_parameters={"query_file": str(query_file)},
            comment="",
        )
        print()
        return add_source

    def create_api_source(self, *, endpoint: str) -> colrev.settings.SearchSource:
        """Interactively add an API SearchSource"""

        print(f"Add {endpoint} as an API SearchSource")
        print()

        keywords = input("Enter the keywords:")

        filename = self.get_unique_filename(
            file_path_string=f"{endpoint.replace('colrev.', '')}"
        )
        add_source = colrev.settings.SearchSource(
            endpoint=endpoint,
            filename=filename,
            search_type=SearchType.API,
            search_parameters={"query": keywords},
            comment="",
        )
        return add_source

    def select_search_type(self, *, search_types: list, params: dict) -> SearchType:
        """Select the SearchType (interactively if neccessary)"""

        if Fields.URL in params:
            return SearchType.API
        if "search_file" in params:
            return SearchType.DB

        choices = [x for x in search_types if x != SearchType.MD]
        if len(choices) == 1:
            return choices[0]
        choices.sort()
        questions = [
            inquirer.List(
                "search_type",
                message="Select SearchType:",
                choices=choices,
            ),
        ]
        answers = inquirer.prompt(questions)
        return SearchType[answers["search_type"]]

    def run_db_search(  # type: ignore
        self,
        *,
        search_source_cls,
        source: colrev.settings.SearchSource,
    ) -> None:
        """Interactively run a DB search"""

        if self.review_manager.in_ci_environment():
            raise colrev_exceptions.SearchNotAutomated("DB search not automated.")

        print("DB search (update)")
        print(
            f"- Go to {Colors.ORANGE}{search_source_cls.db_url}{Colors.END} "
            "and run the following query:"
        )
        print()
        print(f"{Colors.ORANGE}{source.get_query()}{Colors.END}")
        print()
        print(
            f"- Replace search results in {Colors.ORANGE}"
            + str(source.filename)
            + Colors.END
        )
        input("Press enter to continue")
        if source.filename.is_file():
            self.review_manager.dataset.add_changes(source.filename)
        else:
            print("Search results not found.")

    def _get_search_sources(
        self, *, selection_str: typing.Optional[str] = None
    ) -> list[colrev.settings.SearchSource]:
        sources_selected = self.sources
        if selection_str and selection_str != "all":
            selected_filenames = {Path(f).name for f in selection_str.split(",")}
            sources_selected = [
                s for s in self.sources if s.filename.name in selected_filenames
            ]

        assert len(sources_selected) != 0
        return sources_selected

    def _remove_forthcoming(self, source: colrev.settings.SearchSource) -> None:
        """Remove forthcoming papers from a SearchSource"""

        if self.review_manager.settings.search.retrieve_forthcoming:
            return

        if source.filename.suffix != ".bib":
            print(f"{source.filename.suffix} not yet supported")
            return

        records = colrev.loader.load_utils.load(
            filename=source.filename,
            logger=self.review_manager.logger,
        )

        record_list = list(records.values())
        before = len(record_list)
        record_list = [
            r for r in record_list if "forthcoming" != r.get(Fields.YEAR, "")
        ]
        removed = before - len(record_list)
        self.review_manager.logger.info(
            f"{Colors.GREEN}Removed {removed} forthcoming{Colors.END}"
        )
        records = {r[Fields.ID]: r for r in record_list}

        write_file(records_dict=records, filename=source.filename)

    # pylint: disable=no-self-argument
    def _check_source_selection_exists(var_name: str) -> typing.Callable:  # type: ignore
        """Check if the source selection exists"""

        # pylint: disable=no-self-argument
        def check_accepts(func_in: typing.Callable) -> typing.Callable:
            def new_f(self, *args, **kwds) -> typing.Callable:  # type: ignore
                if kwds.get(var_name, None) is None:
                    return func_in(self, *args, **kwds)
                for search_source in kwds[var_name].split(","):
                    if Path(search_source) not in [
                        s.filename for s in self.review_manager.settings.sources
                    ]:
                        raise colrev_exceptions.ParameterError(
                            parameter="select",
                            value=kwds[var_name],
                            options=[
                                str(s.filename)
                                for s in self.review_manager.settings.sources
                            ],
                        )
                return func_in(self, *args, **kwds)

            new_f.__name__ = func_in.__name__
            return new_f

        return check_accepts

    def get_new_search_files(self) -> list[Path]:
        """Retrieve new search files (not yet registered in settings)"""

        search_dir = self.review_manager.paths.search
        files = [
            f.relative_to(self.review_manager.path) for f in search_dir.glob("**/*")
        ]

        # Only files that are not yet registered
        # (also exclude bib files corresponding to a registered file)
        files = [
            f
            for f in files
            if f not in [s.filename for s in self.review_manager.settings.sources]
            and not str(f).endswith("_query.txt")
            and not str(f).endswith(".tmp")
            and ".~lock" not in str(f)
        ]

        return sorted(list(set(files)))

    def _get_heuristics_results_list(
        self,
        *,
        filepath: Path,
        search_sources: dict,
        data: str,
    ) -> list:
        results_list = []
        for endpoint in search_sources:
            search_source_class = self.package_manager.get_package_endpoint_class(
                package_type=EndpointType.search_source,
                package_identifier=endpoint,
            )
            res = search_source_class.heuristic(filepath, data)  # type: ignore
            self.review_manager.logger.debug(f"- {endpoint}: {res['confidence']}")
            if res["confidence"] == 0.0:
                continue
            try:
                result_item = {}

                res["endpoint"] = endpoint

                search_type = SearchType.DB
                # Note : as the identifier, we use the filename
                # (if search results are added by file/not via the API)

                source_candidate = colrev.settings.SearchSource(
                    endpoint=endpoint,
                    filename=filepath,
                    search_type=search_type,
                    search_parameters={},
                    comment="",
                )

                result_item["source_candidate"] = source_candidate
                result_item["confidence"] = res["confidence"]

                results_list.append(result_item)
            except colrev_exceptions.UnsupportedImportFormatError:
                continue
        return results_list

    def _apply_source_heuristics(
        self, *, filepath: Path, search_sources: dict
    ) -> list[typing.Dict]:
        """Apply heuristics to identify source"""

        data = ""
        try:
            data = filepath.read_text()
        except UnicodeDecodeError:
            pass

        results_list = self._get_heuristics_results_list(
            filepath=filepath,
            search_sources=search_sources,
            data=data,
        )

        # Reduce the results_list when there are results with very high confidence
        if [r for r in results_list if r["confidence"] > 0.95]:
            results_list = [r for r in results_list if r["confidence"] > 0.8]

        return results_list

    def add_most_likely_sources(self, *, create_query_files: bool = True) -> None:
        """Get the most likely SearchSources

        returns a dictionary:
        {"filepath": [SearchSource1,..]}
        """

        new_search_files = self.get_new_search_files()
        for filename in new_search_files:
            heuristic_list = self.get_new_source_heuristic(filename)

            for results_list in heuristic_list:
                # Use the last / unknown_source
                max_conf = 0.0
                best_candidate_pos = 0
                for i, heuristic_candidate in enumerate(results_list):
                    if heuristic_candidate["confidence"] > max_conf:
                        best_candidate_pos = i + 1
                        max_conf = heuristic_candidate["confidence"]
                if not any(c["confidence"] > 0.1 for c in results_list):
                    source = [
                        x
                        for x in results_list
                        if x["source_candidate"].endpoint == "colrev.unknown_source"
                    ][0]
                else:
                    selection = str(best_candidate_pos)
                    source = results_list[int(selection) - 1]

                if create_query_files:
                    query_path = self.get_query_filename(
                        filename=source["source_candidate"].filename
                    )
                    source["source_candidate"].search_parameters["query_file"] = str(
                        query_path
                    )
                    query_path.write_text("", encoding="utf-8")
                    self.review_manager.dataset.add_changes(query_path)

                self.review_manager.settings.sources.append(source["source_candidate"])
        self.review_manager.save_settings()
        self.review_manager.dataset.create_commit(msg="Add new search sources")

    def get_new_source_heuristic(self, filename: Path) -> list:
        """Get the heuristic result list of SearchSources candidates

        returns a dictionary:
        {"filepath": ({"search_source": SourceCandidate1", "confidence": 0.98},..]}
        """

        self.review_manager.logger.debug("Load available search_source endpoints...")

        search_sources = self.package_manager.discover_installed_packages(
            package_type=EndpointType.search_source
        )

        heuristic_results = []
        self.review_manager.logger.debug(f"Discover new DB source file: {filename}")

        heuristic_results.append(
            self._apply_source_heuristics(
                filepath=filename,
                search_sources=search_sources,
            )
        )

        return heuristic_results

    def add_source_and_search(
        self, search_source: colrev.settings.SearchSource
    ) -> None:
        """Add a SearchSource and run the search"""

        self.review_manager.settings.sources.append(search_source)
        self.review_manager.save_settings()
        self.review_manager.dataset.create_commit(
            msg=f"Search: add {search_source.endpoint}:{search_source.search_type} → "
            f"data/search/{search_source.filename.name}"
        )
        if not search_source.filename.is_file():
            print()
            self.main(selection_str=str(search_source.filename), rerun=True)

    @_check_source_selection_exists(  # pylint: disable=too-many-function-args
        "selection_str"
    )
    @colrev.process.operation.Operation.decorate()
    def main(
        self,
        *,
        selection_str: typing.Optional[str] = None,
        rerun: bool,
        skip_commit: bool = False,
    ) -> None:
        """Search for records (main entrypoint)"""

        rerun_flag = "" if not rerun else f" ({Colors.GREEN}rerun{Colors.END})"
        self.review_manager.logger.info(f"Search{rerun_flag}")
        self.review_manager.logger.info(
            "Retrieve new records from an API or files (search sources)."
        )
        self.review_manager.logger.info(
            "See https://colrev-environment.github.io/colrev/manual/metadata_retrieval/search.html"
        )

        # Reload the settings because the search sources may have been updated
        self.review_manager.settings = self.review_manager.load_settings()
        for source in self._get_search_sources(selection_str=selection_str):
            try:
                if not self.review_manager.high_level_operation:
                    print()
                self.review_manager.logger.info(
                    f"search: {source.endpoint}:{source.search_type} → "
                    f"data/search/{source.filename.name}"
                )

                search_source_class = self.package_manager.get_package_endpoint_class(
                    package_type=EndpointType.search_source,
                    package_identifier=source.endpoint,
                )
                endpoint = search_source_class(
                    source_operation=self, settings=source.model_dump()
                )

                endpoint.search(rerun=rerun)  # type: ignore

                if not source.filename.is_file():
                    continue

                self._remove_forthcoming(source)
                self.review_manager.dataset.add_changes(source.filename)
                if not skip_commit:
                    self.review_manager.dataset.create_commit(
                        msg=f"Search: run {source.endpoint}:{source.search_type} → "
                        f"data/search/{source.filename.name}"
                    )
            except colrev_exceptions.ServiceNotAvailableException:
                self.review_manager.logger.warning("ServiceNotAvailableException")
            except colrev_exceptions.SearchNotAutomated as exc:
                self.review_manager.logger.warning(exc)
            except colrev_exceptions.MissingDependencyError as exc:
                self.review_manager.logger.warning(exc)

        if self.review_manager.in_ci_environment():
            print("\n\n")



#! /usr/bin/env python
"""Indexing and retrieving records locally."""
from __future__ import annotations

import collections
import io
import os
import typing
from copy import deepcopy
from datetime import timedelta
from multiprocessing import Lock
from pathlib import Path
from threading import Timer

import pandas as pd
import requests_cache
from tqdm import tqdm

import colrev.env.environment_manager
import colrev.env.local_index_sqlite
import colrev.env.resources
import colrev.env.tei_parser
import colrev.env.utils
import colrev.exceptions as colrev_exceptions
import colrev.loader.load_utils
import colrev.ops.check
import colrev.record.record
import colrev.review_manager
from colrev.constants import Colors
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.constants import Filepaths
from colrev.constants import LocalIndexFields
from colrev.constants import RecordState
from colrev.env.local_index_prep import prepare_record_for_indexing
from colrev.writer.write_utils import to_string


class LocalIndexBuilder:
    """The LocalIndexBuilder implements indexing functionality"""

    def __init__(
        self,
        *,
        index_tei: bool = False,
        verbose_mode: bool = False,
    ) -> None:
        self.verbose_mode = verbose_mode
        self.environment_manager = colrev.env.environment_manager.EnvironmentManager()
        self._index_tei = index_tei
        self.thread_lock = Lock()

    def reinitialize_sqlite_db(self) -> None:
        """Reinitialize the SQLITE database ()"""

        Filepaths.LOCAL_INDEX_SQLITE_FILE.unlink(missing_ok=True)
        colrev.env.local_index_sqlite.SQLiteIndexRecord(reinitialize=True)
        colrev.env.local_index_sqlite.SQLiteIndexTOC(reinitialize=True)

    def _outlets_duplicated(self) -> bool:
        print("Validate curated metadata")

        try:
            curated_outlets = self.environment_manager.get_curated_outlets()
            if len(curated_outlets) != len(set(curated_outlets)):
                duplicated = [
                    item
                    for item, count in collections.Counter(curated_outlets).items()
                    if count > 1
                ]
                print(
                    f"Error: Duplicate outlets in curated_metadata : {','.join(duplicated)}"
                )
                return True

        except colrev_exceptions.CuratedOutletNotUnique as exc:
            print(exc)
            return True
        return False

    def _drop_toc_item(
        self, *, toc_to_index: dict, copy_for_toc_index: dict, curated_masterdata: bool
    ) -> None:
        if not curated_masterdata or copy_for_toc_index.get(
            Fields.ENTRYTYPE, ""
        ) not in [
            ENTRYTYPES.ARTICLE,
            ENTRYTYPES.INPROCEEDINGS,
        ]:
            return

        toc_item = colrev.record.record.Record(copy_for_toc_index).get_toc_key()
        # Note : drop (do not index) tocs where records are missing
        # otherwise, record-not-in-toc will be triggered erroneously.
        drop_toc = copy_for_toc_index[
            Fields.STATUS
        ] not in RecordState.get_post_x_states(state=RecordState.md_processed)
        try:
            colrev_id = colrev.record.record.Record(copy_for_toc_index).get_colrev_id(
                assume_complete=True
            )
        except colrev_exceptions.NotEnoughDataToIdentifyException:
            drop_toc = True
        if drop_toc:
            toc_to_index[toc_item] = "DROPPED"
        elif toc_to_index.get("toc_item", "") != "DROPPED":
            if toc_item in toc_to_index:
                toc_to_index[toc_item] += f";{colrev_id}"
            else:
                toc_to_index[toc_item] = colrev_id

    def _add_index_records(self, *, recs_to_index: list, curated_fields: list) -> None:
        list_to_add = [
            {
                k: v
                for k, v in el.items()
                if k in colrev.env.local_index_sqlite.SQLiteIndexRecord.KEYS
            }
            for el in recs_to_index
        ]
        sqlite_index_record = colrev.env.local_index_sqlite.SQLiteIndexRecord()
        for item in list_to_add:
            for (
                records_index_required_key
            ) in colrev.env.local_index_sqlite.SQLiteIndexRecord.KEYS:
                if records_index_required_key not in item:
                    item[records_index_required_key] = ""
            if item[LocalIndexFields.ID] == "":
                print("NO ID IN RECORD")
                continue

            if not sqlite_index_record.exists(local_index_id=item[LocalIndexFields.ID]):
                sqlite_index_record.insert(item)
            else:
                if not curated_fields:
                    continue
                try:
                    stored_record = sqlite_index_record.get(
                        key=Fields.COLREV_ID,
                        value=item[Fields.COLREV_ID],
                    )

                    self._amend_record(
                        sqlite_index_record=sqlite_index_record,
                        stored_record_dict=stored_record,
                        item_to_add=item,
                        curated_fields=curated_fields,
                    )
                except colrev_exceptions.RecordNotInIndexException:  # pragma: no cover
                    pass

        sqlite_index_record.commit()

    def _amend_record(
        self,
        *,
        sqlite_index_record: colrev.env.local_index_sqlite.SQLiteIndexRecord,
        stored_record_dict: dict,
        item_to_add: dict,
        curated_fields: list,
    ) -> None:
        """Adds layered fields to amend existing records"""

        item_record_dict = colrev.loader.load_utils.loads(
            load_string=item_to_add[LocalIndexFields.BIBTEX],
            implementation="bib",
            unique_id_field="ID",
        )
        item_record = colrev.record.record.Record(list(item_record_dict.values())[0])
        stored_record = colrev.record.record.Record(stored_record_dict)

        for curated_field in curated_fields:
            if curated_field in stored_record.data:
                print(f"{curated_field} already in record")
                continue

            if curated_field not in item_record.data:
                continue
            stored_record.update_field(
                key=curated_field,
                value=item_record.data[curated_field],
                source=item_record.get_field_provenance_source(curated_field),
            )

        bibtex = to_string(
            records_dict={stored_record.data[Fields.ID]: stored_record.data},
            implementation="bib",
        )

        sqlite_index_record.update(
            local_index_id=item_to_add[LocalIndexFields.ID], bibtex=bibtex
        )

    # pylint: disable=too-many-arguments
    def index_records(
        self,
        *,
        records: dict,
        repo_source_path: Path,
        curation_url: str,
        curated_masterdata: bool,
        curated_fields: list,
    ) -> None:
        """Index a CoLRev project"""

        recs_to_index = []
        toc_to_index: typing.Dict[str, str] = {}
        for record_dict in tqdm(records.values()):
            copy_for_toc_index = deepcopy(record_dict)
            try:
                record_dict[Fields.METADATA_SOURCE_REPOSITORY_PATHS] = str(
                    repo_source_path
                )
                if curated_fields:
                    for curated_field in curated_fields:
                        colrev.record.record.Record(record_dict).add_field_provenance(
                            key=curated_field, source=f"CURATED:{curation_url}"
                        )
                if curated_masterdata:
                    colrev.record.record.Record(record_dict).set_masterdata_curated(
                        curation_url
                    )

                # Set absolute file paths and set bibtex field (for simpler retrieval)
                if Fields.FILE in record_dict:
                    record_dict.update(
                        file=repo_source_path / Path(record_dict[Fields.FILE])
                    )
                record_dict[LocalIndexFields.BIBTEX] = to_string(
                    records_dict={record_dict[Fields.ID]: record_dict},
                    implementation="bib",
                )
                record_dict = prepare_record_for_indexing(record_dict)
                recs_to_index.append(record_dict)

            except (
                colrev_exceptions.RecordNotIndexableException,
                colrev_exceptions.NotTOCIdentifiableException,
                colrev_exceptions.NotEnoughDataToIdentifyException,
            ) as exc:
                if self.verbose_mode:
                    print(exc)
                    print(record_dict)
            finally:
                self._drop_toc_item(
                    toc_to_index=toc_to_index,
                    copy_for_toc_index=copy_for_toc_index,
                    curated_masterdata=curated_masterdata,
                )
        # Select fields and insert into index (sqlite)

        self._index_tei_document(recs_to_index)

        self._add_index_records(
            recs_to_index=recs_to_index, curated_fields=curated_fields
        )

        if curated_masterdata:
            sqlite_index_toc = colrev.env.local_index_sqlite.SQLiteIndexTOC()
            sqlite_index_toc.add(toc_to_index)

    def _load_masterdata_curations(self) -> dict:  # pragma: no cover
        # Note : the following should be replaced by heuristics
        # based on the data (between colrev load and prep)
        masterdata_curations = {}
        filedata = colrev.env.utils.get_package_file_content(
            module="colrev.env", filename=Path("masterdata_curations.csv")
        )

        if filedata:
            for masterdata_curation in filedata.decode("utf-8").splitlines():
                masterdata_curations[masterdata_curation.lower()] = (
                    masterdata_curation.lower()
                )

        return masterdata_curations

    def index_colrev_project(self, repo_source_path: Path) -> None:  # pragma: no cover
        """Index a CoLRev project"""
        try:
            if not Path(repo_source_path).is_dir():
                print(f"Warning {repo_source_path} not a directory")
                return

            print(f"Index records from {repo_source_path}")
            os.chdir(repo_source_path)
            review_manager = colrev.review_manager.ReviewManager(
                path_str=str(repo_source_path)
            )

            check_operation = colrev.ops.check.CheckOperation(review_manager)

            if review_manager.dataset.get_repo().active_branch.name != "main":
                print(
                    f"{Colors.ORANGE}Warning: {repo_source_path} not on main branch{Colors.END}"
                )

            records_file = check_operation.review_manager.paths.records
            if not records_file.is_file():
                return
            records = check_operation.review_manager.dataset.load_records_dict()

            curation_endpoints = [
                x
                for x in check_operation.review_manager.settings.data.data_package_endpoints
                if x["endpoint"] == "colrev.colrev_curation"
            ]

            curated_fields = []
            curation_url = ""
            if curation_endpoints:
                curation_endpoint = curation_endpoints[0]
                # Set masterdata_provenace to CURATED:{url}
                curation_url = curation_endpoint["curation_url"]
                if (
                    not check_operation.review_manager.settings.is_curated_masterdata_repo()
                ):
                    # Add curation_url to curated fields (provenance)
                    curated_fields = curation_endpoint["curated_fields"]

            curated_masterdata = (
                check_operation.review_manager.settings.is_curated_masterdata_repo()
            )

            self.index_records(
                records=records,
                repo_source_path=repo_source_path,
                curated_fields=curated_fields,
                curation_url=curation_url,
                curated_masterdata=curated_masterdata,
            )

        # TypeErrors are thrown when a repo is in interactive rebase mode
        except (colrev_exceptions.CoLRevException, TypeError) as exc:
            print(exc)

    def index(self) -> None:  # pragma: no cover
        """Index all registered CoLRev projects"""

        # Note : this task takes long and does not need to run often
        session = requests_cache.CachedSession(
            str(Filepaths.PREP_REQUESTS_CACHE_FILE),
            backend="sqlite",
            expire_after=timedelta(days=30),
        )
        # Note : lambda is necessary to prevent immediate function call
        # pylint: disable=unnecessary-lambda
        Timer(0.1, lambda: session.remove_expired_responses()).start()

        if self._outlets_duplicated():
            return

        self.reinitialize_sqlite_db()

        repo_source_paths = [
            x["repo_source_path"] for x in self.environment_manager.local_repos()
        ]
        if not repo_source_paths:
            env_resources = colrev.env.resources.Resources()
            curated_resources = list(self._load_masterdata_curations().values())
            for curated_resource in curated_resources:
                print(f"Install {curated_resource}")
                env_resources.install_curated_resource(
                    curated_resource=curated_resource
                )

            repo_source_paths = [
                x["repo_source_path"] for x in self.environment_manager.local_repos()
            ]

        for repo_source_path in repo_source_paths:
            self.index_colrev_project(repo_source_path)

    def _index_tei_document(self, recs_to_index: list) -> None:
        if not self._index_tei:
            return
        for record_dict in recs_to_index:
            if not Path(record_dict.get(Fields.FILE, "NA")).is_file():
                continue

            try:
                tei_path = self._get_tei_index_file(
                    local_index_id=record_dict[LocalIndexFields.ID]
                )
                tei_path.parents[0].mkdir(exist_ok=True, parents=True)
                if not tei_path.is_file():
                    print(f"Create tei for {record_dict[Fields.FILE]}")
                tei = colrev.env.tei_parser.TEIParser(
                    environment_manager=self.environment_manager,
                    pdf_path=Path(record_dict[Fields.FILE]),
                    tei_path=tei_path,
                )
                record_dict[LocalIndexFields.TEI] = str(tei_path)
                record_dict[Fields.FULLTEXT] = tei.get_tei_str()

            except (
                colrev_exceptions.TEIException,
                AttributeError,
                FileNotFoundError,
                colrev_exceptions.ServiceNotAvailableException,
            ):  # pragma: no cover
                pass

    def _get_tei_index_file(self, *, local_index_id: str) -> Path:
        return Filepaths.TEI_INDEX_DIR / Path(
            f"{local_index_id[:2]}/{local_index_id[2:]}.tei.xml"
        )

    def index_journal_rankings(self) -> None:
        """Indexes journal rankings in sqlite database"""

        filedata = colrev.env.utils.get_package_file_content(
            module="colrev.env", filename=Path("journal_rankings.csv")
        )
        if filedata is not None:
            data_frame = pd.read_csv(
                io.StringIO(filedata.decode("utf-8")), encoding="utf-8"
            )
            sqlite_index_ranking = colrev.env.local_index_sqlite.SQLiteIndexRankings()
            sqlite_index_ranking.insert_df(data_frame)



#! /usr/bin/env python
"""Package init."""
from __future__ import annotations

import inspect
import json
import os
import re
import subprocess
from datetime import datetime
from importlib import import_module
from pathlib import Path
from typing import Type

import git
import inquirer
import pkg_resources
import zope.interface.interface

from colrev.constants import Colors
from colrev.package_manager.interfaces import DataInterface
from colrev.package_manager.interfaces import DedupeInterface
from colrev.package_manager.interfaces import INTERFACE_MAP
from colrev.package_manager.interfaces import PDFGetInterface
from colrev.package_manager.interfaces import PDFGetManInterface
from colrev.package_manager.interfaces import PDFPrepInterface
from colrev.package_manager.interfaces import PDFPrepManInterface
from colrev.package_manager.interfaces import PrepInterface
from colrev.package_manager.interfaces import PrepManInterface
from colrev.package_manager.interfaces import PrescreenInterface
from colrev.package_manager.interfaces import ReviewTypeInterface
from colrev.package_manager.interfaces import ScreenInterface
from colrev.package_manager.interfaces import SearchSourceInterface


def _get_default_author() -> dict:
    try:
        name = (
            subprocess.check_output(["git", "config", "--global", "user.name"])
            .decode()
            .strip()
        )
        email = (
            subprocess.check_output(["git", "config", "--global", "user.email"])
            .decode()
            .strip()
        )
        return {"name": name, "email": email}
    except subprocess.CalledProcessError:
        return {}


DEFAULT_AUTHOR = _get_default_author()

AUTHOR_PATTERN = r"([a-zA-Z ]+)<([a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,})>"


# Function to create the pyproject.toml file
def _create_pyproject_toml(data: dict) -> None:

    plugins = []
    package_location = data["name"]
    if data["built_in"]:
        package_location = data["name"].replace("colrev.", "colrev.packages.")
    for key, value in data["plugins"].items():
        formatted_string = (
            f"{key} = \"{package_location}.src.{value['module']}:{value['class']}\""
        )
        plugins.append(formatted_string)
    plugins_string = "\n".join(plugins)

    content = f"""
[tool.poetry]
name = "{data['name']}"
description = "{data['description']}"
version = "{data['version']}"
license = "{data['license']}"
authors = ["{data['author']['name']} <{data['author']['email']}>"]
repository = "{data['repository']}"

[[tool.poetry.packages]]
include = "src"

[tool.poetry.dependencies]
python = ">=3.9, <4"

[tool.colrev]
colrev_doc_description = "{data['doc_description']}"
colrev_doc_link = "docs/README.md"
search_types = []

[tool.poetry.plugins.colrev]
{plugins_string}

[build-system]
requires = ["poetry-core>=1.0.0", "cython<3.0"]
build-backend = "poetry.core.masonry.api"
"""
    with open("pyproject.toml", "w", encoding="utf-8") as file:
        file.write(content.strip())


def _create_git_repo() -> None:
    repo_path = os.getcwd()
    repo = git.Repo.init(repo_path)
    repo.git.add(all=True)
    repo.index.commit("Initial commit")


def _create_built_in_package() -> bool:
    questions = [
        inquirer.List(
            "package_type",
            message="Select the package type",
            choices=["Standalone", "Built-in"],
        )
    ]
    data = inquirer.prompt(questions)
    return data["package_type"] == "Built-in"


# pylint: disable=unused-argument
def validate_version(answers: list, version: str) -> bool:
    """Validate the version"""
    # return false if the version is not a valid version number
    if version == "":
        return True
    if not re.match(r"^\d+\.\d+\.\d+$", version):
        print(f"  {Colors.RED}The version number is not valid.{Colors.END}")
        return False
    return True


# pylint: disable=unused-argument
def validate_module_name(answers: list, name: str) -> bool:
    """Validate the module name"""
    if name == "":  # Accept default module name
        return True
    # return false if the name is not a valid python module name
    if not re.match(r"^[a-zA-Z_][a-zA-Z0-9_]*$", name):
        print(
            f"  {Colors.RED}The module name is not a valid python module name.{Colors.END}"
        )
        return False
    return True


# pylint: disable=unused-argument
# pylint: disable=too-many-locals
# pylint: disable=too-many-return-statements
# pylint: disable=too-many-branches
# pylint: disable=too-many-statements
def _get_package_data(default_package_name: str, built_in: bool) -> dict:

    default_package_name_str = ""
    if default_package_name:
        default_package_name_str = (
            f" [{Colors.GREEN}{default_package_name}{Colors.END}]"
        )

    default_author_str = (
        f"[{Colors.GREEN}{DEFAULT_AUTHOR['name']} "
        + f"<{DEFAULT_AUTHOR['email']}>{Colors.END}]"
    )

    def validate_author(answers: list, name: str) -> bool:
        if name == "":
            return True
        if not re.match(AUTHOR_PATTERN, name):
            print(
                f"\n  {Colors.RED}The format of the author name and e-mail is not valid."
                "\n It should be like `Max Mustermann <max.mustermann@uni-bamberg.de>`."
                f"{Colors.END}"
            )
            return False
        return True

    def validate_name(answers: list, name: str) -> bool:
        if name == "":  # Accept default package name
            return True

        installed_packages = list(pkg_resources.working_set)
        # Check if the name corresponds to a currently installed python package
        if name in installed_packages:
            print(f"  {Colors.RED}The package name is already installed.{Colors.END}")
            return False

        if built_in:
            if not name.startswith("colrev."):
                print(
                    f"  {Colors.RED}The built-in package name must start with 'colrev.'{Colors.END}"
                )
                return False
            if not re.match(r"^[a-zA-Z_]*\.[a-zA-Z0-9_]*$", name):
                print(
                    f"  {Colors.RED}"
                    "The package name is not a valid python package name."
                    f"{Colors.END}"
                )
                return False
            return True

        # Check if the name is a valid python package name
        if not re.match(r"^[a-zA-Z_][a-zA-Z0-9_]*$", name):
            print(
                f"  {Colors.RED}The package name is not a valid python package name.{Colors.END}"
            )
            return False

        return True

    questions = [
        inquirer.Text(
            "name",
            message=f"Enter the package name, e.g., {default_package_name_str}",
            validate=validate_name,
        ),
        inquirer.Text("description", message="Enter the package description"),
        inquirer.Text(
            "doc_description",
            message="Enter short description string for the CoLRev docs list",
        ),
        inquirer.Text(
            "version",
            message=f"Enter the package version number [{Colors.GREEN}0.1.0{Colors.END}]",
            validate=validate_version,
        ),
        inquirer.Text(
            "license",
            message=f"Enter the package license [{Colors.GREEN}MIT{Colors.END}]",
        ),
        inquirer.Text(
            "author",
            message=f"Enter the author name and e-mail {default_author_str}",
            validate=validate_author,
        ),
    ]
    if not built_in:
        questions.append(
            inquirer.Text("repository", message="Enter the URL of the repository"),
        )

    questions.append(
        inquirer.Checkbox(
            "plugins",
            message="Select the plugin types (use SPACE to select)",
            choices=list(INTERFACE_MAP.keys()),
        )
    )

    package_data = inquirer.prompt(questions)

    def validate_class_name(answers: list, name: str) -> bool:
        if name == "":  # Accept default class name
            return True
        # return false if the name is not a valid python class name
        if not re.match(r"^[a-zA-Z_][a-zA-Z0-9_]*$", name):
            print(
                f"  {Colors.RED}The class name is not a valid python class name.{Colors.END}"
            )
            return False
        return True

    def to_camel_case(snake_str: str) -> str:
        components = snake_str.split("_")
        # Capitalize the first letter of each component and join them together
        return "".join(x.capitalize() for x in components)

    if built_in:
        package_data["repository"] = (
            f"https://github.com/CoLRev-Environment/colrev/tree/"
            f"main/colrev/packages/{package_data['name'].replace('colrev.', '')}"
        )

    if not package_data["name"]:
        package_data["name"] = default_package_name
    if not package_data["version"]:
        package_data["version"] = "0.1.0"
    if not package_data["license"]:
        package_data["license"] = "MIT"
    # pylint: disable=colrev-missed-constant-usage
    if not package_data["author"]:
        package_data["author"] = DEFAULT_AUTHOR
    else:
        match = re.match(AUTHOR_PATTERN, package_data.pop("author"))
        if not match or len(match.groups()) != 2:
            raise ValueError(
                "Author name and e-mail must be in the format 'Name <email>'"
            )
        package_data["author"] = {}
        package_data["author"]["name"] = match.group(1).strip()
        package_data["author"]["email"] = match.group(2)
    if not package_data["repository"]:
        package_data["repository"] = "TODO"

    package_data["doc_description"] = "TODO"

    # Example: package_data["plugins"] = ["data", "screen"]
    plugin_data = {}
    for plugin in package_data["plugins"]:
        p_name = package_data["name"].replace("colrev.", "")
        default_module_name = f"{p_name}_{plugin}"
        default_class_name = to_camel_case(
            f"{p_name.capitalize()}_{plugin.capitalize()}"
        )
        plugin_questions = [
            inquirer.Text(
                "module",
                message=f"Enter the module name for {plugin} "
                + f"[{Colors.GREEN}{default_module_name}{Colors.END}]",
                validate=validate_module_name,
            ),
            inquirer.Text(
                "class",
                message=f"Enter the class name for {plugin} "
                + f"[{Colors.GREEN}{default_class_name}{Colors.END}]",
                validate=validate_class_name,
            ),
        ]

        data = inquirer.prompt(plugin_questions)

        if data["module"] == "":
            data["module"] = default_module_name
        if data["class"] == "":
            data["class"] = default_class_name
        plugin_data[plugin] = data

    package_data["plugins"] = plugin_data
    package_data["built_in"] = built_in

    return package_data


# pylint: disable=too-many-branches
def _get_init_method(interface_class: Type) -> str:

    # pylint: disable=line-too-long
    init_method = """

"""
    if interface_class is ReviewTypeInterface:
        init_method += """    def __init__(
        self, *, operation: colrev.process.operation.Operation, settings: dict
    ) -> None:
        self.settings = colrev.package_manager.package_settings.DefaultSettings(**settings)"""
    elif interface_class is SearchSourceInterface:
        init_method += """
    def __init__(
        self,
        *,
        source_operation: colrev.process.operation.Operation,
        settings: typing.Optional[dict] = None,
    ) -> None:
      pass # TODO"""
    elif interface_class is PrepInterface:
        init_method += """    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = colrev.package_manager.package_settings.DefaultSettings(**settings)"""
    elif interface_class is PrepManInterface:
        init_method += """    def __init__(
        self, *, prep_man_operation: colrev.ops.prep_man.PrepMan, settings: dict
    ) -> None:
        self.settings = colrev.package_manager.package_settings.DefaultSettings(**settings)"""
    elif interface_class is DedupeInterface:
        init_method += """    def __init__(
        self,
        *,
        dedupe_operation: colrev.ops.dedupe.Dedupe,
        settings: dict,
    ) -> None:
      pass # TODO"""
    elif interface_class is PrescreenInterface:
        init_method += """    def __init__(
        self,
        *,
        prescreen_operation: colrev.ops.prescreen.Prescreen,
        settings: dict,
    ) -> None:
      pass # TODO"""
    elif interface_class is PDFGetInterface:
        init_method += """    def __init__(
        self,
        *,
        pdf_get_operation: colrev.ops.pdf_get.PDFGet,
        settings: dict,
    ) -> None:
      pass # TODO"""
    elif interface_class is PDFGetManInterface:
        init_method += """    def __init__(
        self,
        *,
        pdf_get_man_operation: colrev.ops.pdf_get_man.PDFGetMan,
        settings: dict,
    ) -> None:
      pass # TODO"""
    elif interface_class is PDFPrepInterface:
        init_method += """    def __init__(
        self,
        *,
        pdf_prep_operation: colrev.ops.pdf_prep.PDFPrep,
        settings: dict,
    ) -> None:
      pass # TODO"""
    elif interface_class is PDFPrepManInterface:
        init_method += """    def __init__(
        self,
        *,
        pdf_prep_man_operation: colrev.ops.pdf_prep_man.PDFPrepMan,
        settings: dict,
    ) ->: None:
      pass # TODO"""
    elif interface_class is ScreenInterface:
        init_method += """    def __init__(
        self,
        *,
        screen_operation: colrev.ops.screen.Screen,
        settings: dict,
    ) -> None:
      pass # TODO"""
    elif interface_class is DataInterface:
        init_method += """    def __init__(
        self,
        *,
        data_operation: colrev.ops.data.Data,
        settings: dict,
    ) -> None:
      pass # TODO"""
    else:
        init_method += "#TODO"

    return init_method


def _generate_method_signatures(module_path: str, class_name: str) -> list:
    module = import_module(module_path)
    interface_class = getattr(module, class_name)
    members = inspect.getmembers(interface_class)

    interface_class_attrs = next(
        (member for member in members if member[0] == "_InterfaceClass__attrs"), None
    )
    if interface_class_attrs:
        attrs_dict = interface_class_attrs[1]

        method_signatures = []

        for name, attr in attrs_dict.items():
            if isinstance(attr, zope.interface.interface.Attribute) and not isinstance(
                attr, zope.interface.interface.Method
            ):
                method_signatures.append(f"    {name} = '' #TODO\n")
        method_signatures.append("\n")
        method_signatures.append(_get_init_method(interface_class))
        method_signatures.append("\n")
        method_signatures.append("\n")

        for name, attr in attrs_dict.items():
            if isinstance(attr, zope.interface.interface.Method):
                sig_string = attr.getSignatureString().replace("(", "(self, ")
                method_signatures.append(
                    f"    def {name}{sig_string}:\n"
                    f'      """{attr.getDoc()}"""\n      # TODO\n\n'
                )

        return method_signatures

    return ["Attribute '_InterfaceClass__attrs' not found."]


# pylint: disable=too-many-return-statements
# pylint: disable=too-many-branches
def _get_package_imports(plugin: str) -> str:
    if plugin == "review_type":
        return "import colrev.ops.data"
    if plugin == "search_source":
        return "import colrev.process.operation"
    if plugin == "prep":
        return "import colrev.ops.prep"
    if plugin == "prep_man":
        return "import colrev.ops.prep_man"
    if plugin == "dedupe":
        return "import colrev.ops.dedupe"
    if plugin == "prescreen":
        return "import colrev.ops.prescreen"
    if plugin == "pdf_get":
        return "import colrev.ops.pdf_get"
    if plugin == "pdf_get_man":
        return "import colrev.ops.pdf_get_man"
    if plugin == "pdf_prep":
        return "import colrev.ops.pdf_prep"
    if plugin == "pdf_prep_man":
        return "import colrev.ops.pdf_prep_man"
    if plugin == "screen":
        return "import colrev.ops.screen"
    if plugin == "data":
        return "import colrev.ops.data"
    return "#TODO"


def _create_module_files(package_data: dict) -> None:

    os.makedirs("src", exist_ok=True)

    for plugin, data in package_data["plugins"].items():
        file_path = os.path.join("src", f"{data['module']}.py")
        interface = INTERFACE_MAP[plugin]

        module_path = "colrev.package_manager.interfaces"
        method_signatures = _generate_method_signatures(module_path, interface)

        with open(file_path, "w", encoding="utf-8") as file:
            file.write(
                f'''#! /usr/bin/env python
"""{interface}: {data['class']}"""

import typing
from zope.interface import implementer
from colrev.package_manager.interfaces import {interface}
import colrev.package_manager.package_settings
{_get_package_imports(plugin)}

@implementer({interface})
class {data['class']}:
'''
            )
            for signature in method_signatures:
                file.write(signature)


def _create_src_init(package_data: dict) -> None:

    with open("src/__init__.py", "w", encoding="utf-8") as file:
        file.write(
            f'''"""Package for {package_data['name']}."""

__author__ = "{package_data['author']['name']}"
__email__ = "{package_data['author']['email']}"'''
        )


def _create_docs_readme() -> None:

    Path("docs").mkdir()

    with open("docs/README.md", "w", encoding="utf-8") as file:
        file.write("""# TODO : Docs""")


def _create_readme(package_data: dict) -> None:

    with open("README.md", "w", encoding="utf-8") as file:
        file.write(f"""# {package_data['name']}""")
        if package_data["description"]:
            file.write(f"\n\n{package_data['description']}")
        file.write("\n\n## Installation")
        file.write("\n\n```bash")
        file.write(f"\ncolrev install {package_data['name']}")
        file.write("\n```")
        file.write("\n\n## Usage")
        file.write("\n\nTODO")
        file.write("\n\n## License")
        file.write(
            f"\n\nThis project is licensed under the {package_data['license']} "
            "License - see the [LICENSE](LICENSE) file for details."
        )


def _create_license_file(package_data: dict) -> None:

    license_file_path = "LICENSE"
    license_text = _get_license_text(package_data)
    with open(license_file_path, "w", encoding="utf-8") as file:
        file.write(license_text)


def _get_license_text(package_data: dict) -> str:
    license_name = package_data["license"]

    current_year = datetime.now().year
    license_texts = {
        "MIT": f"""MIT License

        Copyright (c) [{current_year}] [{package_data['author']['name']}]

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.""",
        # Add more license texts for other licenses if needed
    }

    license_text = license_texts.get(license_name, "")

    return license_text


def _create_pre_commit_hooks() -> None:

    pre_commit_hooks = """
repos:
-   repo: https://github.com/pre-commit/pre-commit-hooks
    rev: v4.6.0
    hooks:
    -   id: trailing-whitespace
    -   id: end-of-file-fixer
        exclude: bib$|txt$|ris$|enl$|xml$
    -   id: check-docstring-first
    -   id: check-json
    -   id: check-yaml
    -   id: check-toml
    -   id: debug-statements
    -   id: name-tests-test
-   repo: https://github.com/psf/black-pre-commit-mirror
    rev: 24.4.2
    hooks:
      - id: black
        language_version: python3
-   repo: https://github.com/PyCQA/autoflake
    rev: v2.3.1
    hooks:
    -   id: autoflake
-   repo: https://github.com/PyCQA/flake8
    rev: 7.1.0
    hooks:
    -   id: flake8
        additional_dependencies: [flake8-typing-imports==1.12.0]
        args: ['--max-line-length=110', '--extend-ignore=E203,TYP006']
-   repo: https://github.com/asottile/reorder-python-imports
    rev: v3.13.0
    hooks:
    -   id: reorder-python-imports
        args: [--py3-plus]
-   repo: https://github.com/asottile/pyupgrade
    rev: v3.16.0
    hooks:
    -   id: pyupgrade
        args: [--py36-plus, --keep-runtime-typing]
-   repo: https://github.com/pre-commit/mirrors-mypy
    rev: 'v1.11.0'
    hooks:
    -   id: mypy
        args: [--disallow-untyped-defs, --disallow-incomplete-defs, --disallow-untyped-calls]
        additional_dependencies: [types-toml]
- repo: https://github.com/astral-sh/ruff-pre-commit
  rev: v0.5.4
  hooks:
    - id: ruff # runs faster than pylint
      args: [--fix, --exit-non-zero-on-fix]
- repo: local
  hooks:
    - id: pylint
      name: pylint
      entry: pylint
      language: system
      types: [python]
      files: colrev
      args:
        [
          "-rn", # Only display messages
          "-sn", # Don't display the score
        ]
"""
    with open(".pre-commit-config.yaml", "w", encoding="utf-8") as file:
        file.write(pre_commit_hooks)


def _add_to_packages_json(package_data: dict) -> None:
    packages_json_path = (
        Path(__file__).resolve().parent.parent.parent
        / "colrev/package_manager/packages.json"
    )
    with open(packages_json_path, encoding="utf-8") as file:

        data = json.load(file)

        # Add the package with "experimental" status
        data[package_data["name"]] = {"dev_status": "experimental"}

        with open(packages_json_path, "w", encoding="utf-8") as file_out:
            json.dump(data, file_out, indent=4)


def main() -> None:
    """Main function to create a CoLRev package."""

    built_in = _create_built_in_package()
    current_dir = Path.cwd()
    if any(Path(current_dir).iterdir()):
        print("The current directory is not empty. Please select an empty directory.")
        return
    colrev_packages_dir = Path(__file__).resolve().parent.parent / "packages"
    if built_in:
        if not colrev_packages_dir.is_dir() or not str(current_dir).startswith(
            str(colrev_packages_dir)
        ):
            print(
                "To create a built-in package, "
                "please navigate to a subdirectory of the 'packages' directory:"
                f"\n{colrev_packages_dir}/{Colors.ORANGE}<your-package_name>{Colors.END}"
            )
            return
        if current_dir.name.startswith("colrev."):
            print("The built-in package dir should not start with 'colrev.'")
            return
        default_package_name = f"colrev.{current_dir.name}"

    else:
        if str(current_dir).startswith(str(colrev_packages_dir)):
            print(
                "To create a standalone package, "
                "please navigate to a directory outside of the CoLRev package."
            )
            return

        default_package_name = current_dir.name

    package_data = _get_package_data(default_package_name, built_in)

    _create_pyproject_toml(package_data)
    _create_module_files(package_data)
    _create_src_init(package_data)
    _create_readme(package_data)
    _create_license_file(package_data)
    _create_pre_commit_hooks()
    _create_docs_readme()
    if not built_in:
        _create_git_repo()

    if built_in:
        _add_to_packages_json(package_data)

    # tests
    # gh-actions: pypi-publishing setup



#! /usr/bin/env python
"""Checker for pdf-incomplete."""
from __future__ import annotations

import re
from pathlib import Path

import colrev.env.utils
import colrev.record.qm.quality_model
from colrev.constants import Fields
from colrev.constants import PDFDefectCodes

# pylint: disable=too-few-public-methods

FULL_VERSION_PURCHASE_NOTICES = [
    "morepagesareavailableinthefullversionofthisdocument,whichmaybepurchas"
]
ROMAN_PAGES_PATTERN = re.compile(
    r"^M{0,3}(CM|CD|D?C{0,3})?(XC|XL|L?X{0,3})?(IX|IV|V?I{0,3})?--"
    + r"M{0,3}(CM|CD|D?C{0,3})?(XC|XL|L?X{0,3})?(IX|IV|V?I{0,3})?$",
    re.IGNORECASE,
)
ROMAN_PAGE_PATTERN = re.compile(
    r"^M{0,3}(CM|CD|D?C{0,3})?(XC|XL|L?X{0,3})?(IX|IV|V?I{0,3})?$", re.IGNORECASE
)


class PDFIncompletenessChecker:
    """The PDFIncompletenessChecker"""

    msg = PDFDefectCodes.PDF_INCOMPLETE

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.record_pdf.PDFRecord) -> None:
        """Run the pdf-incomplete checks"""

        if (
            Fields.FILE not in record.data
            or Path(record.data[Fields.FILE]).suffix != ".pdf"
            or Fields.PAGES not in record.data
            or record.ignored_defect(key=Fields.FILE, defect=self.msg)
        ):
            return

        if not self._pages_match_pdf(record=record):
            record.add_field_provenance_note(key=Fields.FILE, note=self.msg)
        else:
            record.remove_field_provenance_note(key=Fields.FILE, note=self.msg)

    def _pages_match_pdf(self, *, record: colrev.record.record_pdf.PDFRecord) -> bool:

        def longer_with_appendix(
            *,
            record: colrev.record.record_pdf.PDFRecord,
            nr_pages: int,
        ) -> bool:
            if 10 < nr_pages < record.data[Fields.NR_PAGES_IN_FILE]:
                text = record.extract_text_by_page(
                    pages=list(
                        range(nr_pages + 1, record.data[Fields.NR_PAGES_IN_FILE] + 1)
                    )
                )
                if "appendi" in text.lower():
                    return True
            return False

        def roman_to_int(input_str: str) -> int:
            input_str = input_str.lower()
            roman = {
                "i": 1,
                "v": 5,
                "x": 10,
                "l": 50,
                "c": 100,
                "d": 500,
                "m": 1000,
                "iv": 4,
                "ix": 9,
                "xl": 40,
                "xc": 90,
                "cd": 400,
                "cm": 900,
            }
            i = 0
            num = 0
            while i < len(input_str):
                if i + 1 < len(input_str) and input_str[i : i + 2] in roman:
                    num += roman[input_str[i : i + 2]]
                    i += 2
                else:
                    num += roman[input_str[i]]
                    i += 1
            return num

        def get_nr_pages(*, pages: str) -> int:
            pages_str = pages

            roman_pages_matched = re.match(ROMAN_PAGES_PATTERN, pages)
            if roman_pages_matched:
                start_page, end_page = map(
                    roman_to_int, roman_pages_matched.group().split("--")
                )
                pages_str = f"{start_page}--{end_page}"

            roman_page_matched = re.match(ROMAN_PAGE_PATTERN, pages)
            if roman_page_matched:
                page = roman_page_matched.group()
                pages_str = f"{roman_to_int(page)}"

            if "--" in pages_str:
                start_page, end_page = map(int, pages_str.split("--"))
                nr_pages = end_page - start_page + 1
            else:
                nr_pages = 1
            return nr_pages

        # Get nr pages from PDF (set in quality_model)
        if Fields.NR_PAGES_IN_FILE not in record.data:
            return False

        # Not complete if there is a FULL_VERSION_PURCHASE_NOTICE
        if any(
            FULL_VERSION_PURCHASE_NOTICE
            in record.data[Fields.TEXT_FROM_PDF].lower().replace(" ", "")
            for FULL_VERSION_PURCHASE_NOTICE in FULL_VERSION_PURCHASE_NOTICES
        ):
            return False

        # Get nr pages from pages field
        try:
            nr_pages = get_nr_pages(pages=record.data[Fields.PAGES])
        except ValueError:
            # e.g., S49--S50
            return True

        # Special case: if the PDF has more pages than the pages field, it may be complete
        if longer_with_appendix(record=record, nr_pages=nr_pages):
            return True

        # If the PDF has the same number of pages as the pages field, it is complete
        return nr_pages == record.data[Fields.NR_PAGES_IN_FILE]


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(PDFIncompletenessChecker(quality_model))



#! /usr/bin/env python
"""Retrieval of PDFs from the website (URL)"""
from __future__ import annotations

from pathlib import Path
from urllib.parse import urljoin

import requests
import zope.interface
from bs4 import BeautifulSoup
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields

# pylint: disable=duplicate-code
# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.package_manager.interfaces.PDFGetInterface)
class WebsiteDownload:
    """Get PDFs from the website"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings
    ci_supported: bool = Field(default=False)

    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
        + "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3",
        "Accept-Language": "en-US,en;q=0.9",
        "Accept-Encoding": "gzip, deflate, br",
        "Connection": "keep-alive",
        "Upgrade-Insecure-Requests": "1",
        "TE": "Trailers",
    }

    def __init__(
        self,
        *,
        pdf_get_operation: colrev.ops.pdf_get.PDFGet,
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)
        self.review_manager = pdf_get_operation.review_manager
        self.pdf_get_operation = pdf_get_operation

    def _download_from_jmir(
        self, *, record: colrev.record.record.Record, pdf_filepath: Path
    ) -> None:
        article_url = record.data[Fields.URL]
        response = requests.get(article_url, headers=self.headers, timeout=60)

        if response.status_code == 200:
            soup = BeautifulSoup(response.text, "html.parser")

            pdf_link = soup.find("a", {"data-test": "pdf-button"})

            if pdf_link:
                pdf_url = urljoin(article_url, pdf_link["href"])

                paper_title_tag = soup.find("meta", {"name": "citation_title"})
                if paper_title_tag:
                    pdf_response = requests.get(pdf_url, timeout=60)

                    if pdf_response.status_code == 200:
                        with open(pdf_filepath, "wb") as pdf_file:
                            pdf_file.write(pdf_response.content)
                        self.review_manager.logger.debug(
                            f"PDF downloaded successfully as {pdf_filepath}"
                        )
                    else:
                        self.review_manager.logger.debug(
                            f"Failed to download PDF. Status code: {pdf_response.status_code}"
                        )
                else:
                    self.review_manager.logger.debug(
                        "Paper title not found on the page."
                    )
            else:
                self.review_manager.logger.debug("PDF link not found on the page.")
        else:
            self.review_manager.logger.debug(
                f"Failed to retrieve the article page. Status code: {response.status_code}"
            )

    def _download_from_bmj_open_science(
        self, *, record: colrev.record.record.Record, pdf_filepath: Path
    ) -> None:
        url = record.data[Fields.URL]
        response = requests.get(url, headers=self.headers, timeout=60)

        if response.status_code == 200:
            soup = BeautifulSoup(response.text, "html.parser")

            pdf_link = soup.find("a", {"class": "article-pdf-download"})

            if pdf_link:
                pdf_url = pdf_link.get("href")

                if pdf_url:
                    if not pdf_url.startswith(("http:", "https:")):
                        pdf_url = urljoin(url, pdf_url)

                    pdf_response = requests.get(pdf_url, timeout=60)

                    if pdf_response.status_code == 200:
                        with open(pdf_filepath, "wb") as pdf_file:
                            pdf_file.write(pdf_response.content)

                        self.review_manager.logger.debug(
                            f"PDF downloaded successfully as {pdf_filepath}"
                        )
                    else:
                        self.review_manager.logger.debug(
                            f"Failed to download PDF. Status code: {pdf_response.status_code}"
                        )
                else:
                    self.review_manager.logger.debug("PDF URL not found on the page.")
            else:
                self.review_manager.logger.debug("PDF link not found on the page.")
        else:
            self.review_manager.logger.debug(
                f"Failed to retrieve the article page. Status code: {response.status_code}"
            )

    def _download_unknown(
        self, *, record: colrev.record.record.Record, pdf_filepath: Path
    ) -> None:
        url = record.data[Fields.URL]
        response = requests.get(url, headers=self.headers, timeout=60)

        if response.status_code == 200:
            soup = BeautifulSoup(response.text, "html.parser")

            pdf_link = soup.find("a", text="PDF")
            if not pdf_link:
                pdf_link = soup.find(
                    "a", href=lambda href: href and href.endswith(".pdf")
                )

            if pdf_link:
                pdf_url = pdf_link.get("href")

                if pdf_url:
                    if not pdf_url.startswith(("http:", "https:")):
                        pdf_url = urljoin(url, pdf_url)

                    pdf_response = requests.get(pdf_url, timeout=60)

                    if pdf_response.status_code == 200:
                        with open(pdf_filepath, "wb") as pdf_file:
                            pdf_file.write(pdf_response.content)

                        self.review_manager.logger.debug(
                            f"PDF downloaded successfully as {pdf_filepath}"
                        )
                    else:
                        self.review_manager.logger.debug(
                            f"Failed to download PDF. Status code: {pdf_response.status_code}"
                        )
                else:
                    self.review_manager.logger.debug("PDF URL not found on the page.")
            else:
                self.review_manager.logger.debug("PDF link not found on the page.")
        else:
            self.review_manager.logger.debug(
                f"Failed to retrieve the Minerva Medica page. Status code: {response.status_code}"
            )

    def get_pdf(
        self, record: colrev.record.record.Record
    ) -> colrev.record.record.Record:
        """Get PDFs from website (URL)"""

        if Fields.URL not in record.data:
            return record

        pdf_filepath = self.pdf_get_operation.get_target_filepath(record)

        try:
            if any(
                x in record.data[Fields.URL]
                for x in ["jmir.org", "iproc.org", "researchprotocols.org"]
            ):
                self._download_from_jmir(record=record, pdf_filepath=pdf_filepath)

            if "bmjopensem.bmj.com" in record.data[Fields.URL]:
                self._download_from_bmj_open_science(
                    record=record, pdf_filepath=pdf_filepath
                )

            else:
                self._download_unknown(record=record, pdf_filepath=pdf_filepath)

            if pdf_filepath.is_file():
                record.update_field(
                    key=Fields.FILE, value=str(pdf_filepath), source="website-download"
                )
                self.pdf_get_operation.import_pdf(record)
        except requests.RequestException:
            pass

        return record



#! /usr/bin/env python
"""Package check."""
from __future__ import annotations

import subprocess
import sys
from importlib import import_module
from importlib import util
from pathlib import Path
from typing import Any
from typing import Dict

import toml
from zope.interface.verify import verifyClass

import colrev.package_manager.interfaces
from colrev.constants import Colors
from colrev.package_manager.interfaces import INTERFACE_MAP


def _check_package_installed(data: dict) -> bool:
    package_name = data["tool"]["poetry"]["name"]
    try:
        subprocess.check_output(["pip", "show", package_name])
    except subprocess.CalledProcessError:
        print(
            f"Navigate to {Path.cwd()} and run: {Colors.GREEN} pip install -e .{Colors.END}"
        )

    return True


def _check_key_exists(data: Dict[str, Any], key: str) -> bool:
    keys = key.split(".")
    sub_data = data
    for k in keys:
        if k not in sub_data:
            return False
        sub_data = sub_data[k]
    return True


def _check_tool_poetry(data: dict) -> bool:
    return _check_key_exists(data, "tool.poetry")


def _check_tool_poetry_name(data: dict) -> bool:
    return _check_key_exists(data, "tool.poetry.name")


def _check_tool_poetry_description(data: dict) -> bool:
    return _check_key_exists(data, "tool.poetry.description")


def _check_tool_poetry_version(data: dict) -> bool:
    return _check_key_exists(data, "tool.poetry.version")


def _check_tool_poetry_license(data: dict) -> bool:
    return _check_key_exists(data, "tool.poetry.license")


def _check_tool_poetry_authors(data: dict) -> bool:
    return _check_key_exists(data, "tool.poetry.authors")


def _check_tool_poetry_plugins_colrev(data: dict) -> bool:
    return _check_key_exists(data, "tool.poetry.plugins.colrev")


def _check_tool_poetry_plugins_colrev_keys(data: dict) -> bool:
    colrev_data = (
        data.get("tool", {}).get("poetry", {}).get("plugins", {}).get("colrev", {})
    )
    return all(key in list(INTERFACE_MAP) for key in list(colrev_data))


# pylint: disable=too-many-locals
def _check_tool_poetry_plugins_colrev_classes(data: dict) -> bool:
    colrev_data = (
        data.get("tool", {}).get("poetry", {}).get("plugins", {}).get("colrev", {})
    )

    for interface_identifier, endpoint_str in colrev_data.items():
        module_path, class_name = endpoint_str.rsplit(":")

        package, module_path = module_path.split(".", 1)

        try:

            package_spec = util.find_spec(package)
            if not package_spec:
                return False
            if not package_spec.origin:
                return False
            package_path = Path(package_spec.origin).parent
            filepath = package_path / Path(module_path.replace(".", "/") + ".py")
            module_spec = util.spec_from_file_location(
                f"{package}.{module_path}", filepath
            )

            module = import_module(f"{package}.{module_path}")
            module_spec.loader.exec_module(module)  # type: ignore
            cls = getattr(module, class_name)  # type: ignore
            interface: str = INTERFACE_MAP.get(interface_identifier)  # type: ignore

            interface_class = getattr(colrev.package_manager.interfaces, interface)
            if not interface or not verifyClass(interface_class, cls):
                return False
        except (ImportError, AttributeError) as exc:
            print(exc)
            return False
    return True


def _check_build_system(data: dict) -> bool:
    return _check_key_exists(data, "build-system")


# Define checks with preconditions
checks = {
    "check_tool_poetry": {"method": _check_tool_poetry, "preconditions": []},
    "check_tool_poetry_name": {
        "method": _check_tool_poetry_name,
        "preconditions": ["_check_tool_poetry"],
    },
    "check_tool_poetry_description": {
        "method": _check_tool_poetry_description,
        "preconditions": ["_check_tool_poetry"],
    },
    "check_tool_poetry_version": {
        "method": _check_tool_poetry_version,
        "preconditions": ["_check_tool_poetry"],
    },
    "check_tool_poetry_license": {
        "method": _check_tool_poetry_license,
        "preconditions": ["_check_tool_poetry"],
    },
    "check_tool_poetry_authors": {
        "method": _check_tool_poetry_authors,
        "preconditions": ["_check_tool_poetry"],
    },
    "check_tool_poetry_plugins_colrev": {
        "method": _check_tool_poetry_plugins_colrev,
        "preconditions": ["_check_tool_poetry"],
    },
    "check_tool_poetry_plugins_colrev_keys": {
        "method": _check_tool_poetry_plugins_colrev_keys,
        "preconditions": ["_check_tool_poetry_plugins_colrev"],
    },
    "check_tool_poetry_plugins_colrev_classes": {
        "method": _check_tool_poetry_plugins_colrev_classes,
        "preconditions": ["_check_tool_poetry_plugins_colrev"],
    },
    "check_package_installed": {
        "method": _check_package_installed,
        "preconditions": [],
    },
    "check_build_system": {"method": _check_build_system, "preconditions": []},
}


def _validate_structure(data: dict, checks_dict: dict) -> list:
    failed_checks = []
    for check_name, check in checks_dict.items():
        precondition_failed = any(
            precondition in failed_checks for precondition in check["preconditions"]
        )
        if precondition_failed:
            print(f"Skipping {check_name} due to failed preconditions.")
            continue

        method = check["method"]
        result = method(data)

        if not result:
            print(f"Check failed: {check_name}")
            failed_checks.append(check_name)
        else:
            print(f"Check passed: {check_name}")

    return failed_checks


def main() -> None:
    """Run checks of a CoLRev project."""

    file_path = "pyproject.toml"

    try:
        with open(file_path, encoding="utf-8") as file:
            data = toml.load(file)

        failed_checks = _validate_structure(data, checks)
        if failed_checks:
            print("The pyproject.toml file structure is invalid. Failed checks:")
            for check in failed_checks:
                print(f" - {check}")
        else:
            print("Check passed: check_pyproject_valid_structure")
    except Exception as exc:  # pylint: disable=broad-except
        print(f"Error reading pyproject.toml: {exc}")
        sys.exit(1)



#! /usr/bin/env python
"""Connector for pandas"""
from __future__ import annotations

import os
import shutil
import typing
from pathlib import Path

import pandas as pd
from bib_dedupe.bib_dedupe import block
from bib_dedupe.bib_dedupe import match
from bib_dedupe.bib_dedupe import prep

import colrev.ops.check
import colrev.review_manager
from colrev.constants import Fields
from colrev.constants import RecordState


def load_df(
    *,
    project_path: str = "",
    add_abstracts_from_pdfs: bool = False,
    include_provenance: bool = True,
    notify: str = "",
) -> pd.DataFrame:
    """Get a pandas dataframe from a CoLRev project"""

    if project_path == "":
        project_path = str(Path.cwd())

    os.chdir(project_path)

    if add_abstracts_from_pdfs:
        raise NotImplementedError

    if not include_provenance:
        raise NotImplementedError

    if notify != "":
        raise NotImplementedError

    review_manager = colrev.review_manager.ReviewManager(path_str=project_path)
    colrev.ops.check.CheckOperation(review_manager)
    records = review_manager.dataset.load_records_dict()
    loaded_df = pd.DataFrame.from_dict(records, orient="index")
    return loaded_df


def load_resolved_papers(other_project_path: str) -> pd.DataFrame:
    """Loads records from the other project, matches them with the current project,
    and returns a DataFrame of the other records with the corresponding IDs in the current project.

    This is useful when integrating datasets and reusing screening decisions or coding.

    Example:

    Current project record(s):

    @Webster2002{
        title = "Analyzing the past to prepare for the future",
        author = "Webser, Jane and Watson, Rick",
        ...
    }

    Other project record(s):

    @ID1{
        title = "ANALYZING THE PAST TO PREPARE FOR THE FUTURE",
        author = "WEBSTER; WATSON",
        colrev_status = "rev_included",
        methods = "literature_review",
        ...
    }

    Returns:

    @Webster2002{
        title = "ANALYZING THE PAST TO PREPARE FOR THE FUTURE",
        author = "WEBSTER; WATSON",
        colrev_status = "rev_included",
        methods = "literature_review",
        OTHER_ID = "ID1",
        ...
    }

    Data can easily be assigned to the current project
    by accessing other records based on corresponding current IDs.

    """

    other_review_manager = colrev.review_manager.ReviewManager(
        path_str=other_project_path, force_mode=True
    )
    colrev.ops.check.CheckOperation(other_review_manager)
    # pylint: disable=colrev-records-variable-naming-convention
    other_records = other_review_manager.dataset.load_records_dict()
    project_path = str(Path.cwd())
    review_manager = colrev.review_manager.ReviewManager(
        path_str=project_path, force_mode=True
    )
    colrev.ops.check.CheckOperation(review_manager)
    records = review_manager.dataset.load_records_dict()

    # identify duplicates with dedupe

    records_df = pd.DataFrame.from_dict(records, orient="index")
    records_df["search_set"] = "ORIGINAL"
    records_df[Fields.ID] = "ORIGINAL_" + records_df[Fields.ID].astype(str)
    other_records_df = pd.DataFrame.from_dict(other_records, orient="index")
    other_records_df["search_set"] = "OTHER"
    other_records_df.reset_index(inplace=True)

    combined_df = pd.concat([records_df, other_records_df], ignore_index=True)

    selected_columns = [
        Fields.ID,
        "search_set",
        Fields.AUTHOR,
        Fields.TITLE,
        Fields.YEAR,
        Fields.JOURNAL,
        Fields.BOOKTITLE,
        Fields.VOLUME,
        Fields.NUMBER,
        Fields.PAGES,
        Fields.DOI,
        Fields.ENTRYTYPE,
        Fields.ABSTRACT,
    ]
    combined_df = combined_df[selected_columns]

    combined_df = prep(records_df=combined_df)

    deduplication_pairs = block(combined_df)
    matched_df = match(deduplication_pairs)

    matched_df.rename(columns={"ID_2": "ID"}, inplace=True)
    matched_df = matched_df[matched_df["duplicate_label"] == "duplicate"]

    merged_df = pd.merge(other_records_df, matched_df, on="ID", how="inner")

    merged_df["other_ID"] = merged_df["ID"]
    merged_df[Fields.ID] = merged_df["ID_1"].str.replace("ORIGINAL_", "")
    merged_df.drop(columns=["ID_1"], inplace=True)
    merged_df.drop(columns=["search_set", "duplicate_label", "index"], inplace=True)

    return merged_df


def add_from_tei(
    records_df: pd.DataFrame,
    *,
    project_path: str = "",
    fields: typing.Optional[typing.List[str]] = None,
) -> None:
    """
    This function adds data from TEI files to the given DataFrame.

    Parameters:
    records_df (pd.DataFrame): The DataFrame to which the data will be added.
    project_path (str, optional): The path to the project. Defaults to the
                                  current working directory.
    fields (Optional[List[str]], optional): The fields to be added from the TEI
                                            files. Defaults to [Fields.ABSTRACT,
                                            Fields.KEYWORDS].

    Returns:
    None
    """
    if project_path == "":
        project_path = str(Path.cwd())
    if fields is None:
        fields = [Fields.ABSTRACT, Fields.KEYWORDS]

    review_manager = colrev.review_manager.ReviewManager(
        path_str=project_path, force_mode=True
    )

    def extract_abstract(record: pd.Series) -> str:
        if Fields.ABSTRACT in record and not pd.isnull(record[Fields.ABSTRACT]):
            return record[Fields.ABSTRACT]

        try:
            tei_filename = colrev.record.record.Record(record).get_tei_filename()
            tei = review_manager.get_tei(tei_path=tei_filename)
            return tei.get_abstract()
        except FileNotFoundError:
            return ""

    def extract_keywords(record: pd.Series) -> str:
        if Fields.KEYWORDS in record and not pd.isnull(record[Fields.KEYWORDS]):
            return record[Fields.KEYWORDS]
        try:
            tei_filename = colrev.record.record.Record(record).get_tei_filename()
            tei = review_manager.get_tei(tei_path=tei_filename)
            return ", ".join(tei.get_paper_keywords())
        except FileNotFoundError:
            return ""

    if Fields.ABSTRACT in fields:
        records_df[Fields.ABSTRACT] = records_df.apply(extract_abstract, axis=1)
    if Fields.KEYWORDS in fields:
        records_df[Fields.KEYWORDS] = records_df.apply(extract_keywords, axis=1)


def extract_pdfs_for_data_extraction(
    records_df: pd.DataFrame, directory: str, copy_files: bool = False
) -> None:
    """
    This function creates symlinks or copies the PDFs to the given directory
    based on the copy_files parameter.

    Parameters:
    records_df (pd.DataFrame): The DataFrame containing the records.
    directory (str): The directory where the symlinks or copies will be created.
    copy_files (bool, optional): If True, copies the PDFs instead of creating
                                 symlinks. Defaults to False.

    Returns:
    None
    """
    project_path = str(Path.cwd())
    review_manager = colrev.review_manager.ReviewManager(
        path_str=project_path, force_mode=True
    )

    directory_path = Path(directory)
    directory_path.mkdir(parents=True, exist_ok=True)
    for _, record in records_df.iterrows():
        if record[Fields.STATUS] not in [
            RecordState.rev_included,
            RecordState.rev_synthesized,
        ]:
            continue
        if Fields.FILE in record and not pd.isnull(record[Fields.FILE]):
            pdf_path = review_manager.path / Path(record[Fields.FILE])
            target_path = directory_path / pdf_path.name
            if copy_files:
                if not target_path.exists():
                    shutil.copy(pdf_path, target_path)
            else:
                if not target_path.exists():
                    target_path.symlink_to(pdf_path)


def save(records_df: pd.DataFrame) -> None:
    """
    This function saves the records from a DataFrame to a dataset.

    Parameters:
    records_df (pd.DataFrame): The DataFrame containing the records to be saved.

    Returns:
    None
    """
    project_path = str(Path.cwd())
    review_manager = colrev.review_manager.ReviewManager(
        path_str=project_path, force_mode=True
    )
    records = records_df.to_dict(orient="index")
    for rec_id, record in records.items():
        for key in list(record.keys()):
            if isinstance(record[key], (list, dict)):
                continue
            if pd.isnull(record[key]) or record[key] == "":
                records[rec_id].pop(key)

    review_manager.dataset.save_records_dict(records)



#! /usr/bin/env python
"""Template for a custom PDFGet PackageEndpoint"""
from __future__ import annotations

import zope.interface

import colrev.package_manager.interfaces
import colrev.package_manager.package_settings
import colrev.process.operation
from colrev.constants import Fields


# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.package_manager.interfaces.PDFGetInterface)
class CustomPDFGet:
    """Class for custom pdf-get scripts"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings

    def __init__(
        self,
        *,
        pdf_get_operation: colrev.ops.pdf_get.PDFGet,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)

    def get_pdf(
        self,
        pdf_get_operation: colrev.ops.pdf_get.PDFGet,  # pylint: disable=unused-argument
        record: colrev.record.record.Record,
    ) -> colrev.record.record.Record:
        """Get the PDF"""

        record.data[Fields.FILE] = "filepath"
        pdf_get_operation.import_pdf(record)

        return record
